# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2014-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import os
import time
import json
import shutil
import traceback

from subprocess import check_call, CalledProcessError

from bigml.api import check_resource

from bigmler.processing.models import MONTECARLO_FACTOR
from bigmler.checkpoint import file_number_of_lines
from bigmler.utils import storage_file_name, open_mode
from bigmler.tests.world import world, res_filename, ok_, eq_
from bigmler.tests.ml_tst_prediction_steps import \
    i_create_all_ml_resources
from bigmler.tests.ml_tst_prediction_steps import \
    i_create_all_ml_resources_and_ensembles
from bigmler.tests.ml_tst_evaluation_steps import \
    i_create_all_ml_resources_for_evaluation
from bigmler.tests.max_categories_tst_prediction_steps import \
    i_check_create_max_categories_datasets, i_create_all_mc_resources
from bigmler.tests.basic_batch_tst_prediction_steps import \
    i_check_create_test_source, i_check_create_test_dataset, \
    i_check_create_batch_prediction
from bigmler.tests.basic_cluster_prediction_steps import \
    i_create_all_cluster_resources, i_check_create_cluster
from bigmler.tests.basic_anomaly_prediction_steps import \
    i_create_all_anomaly_resources, i_check_create_anomaly
from bigmler.tests.ml_tst_prediction_steps import \
    i_create_all_mlm_resources
from bigmler.tests.basic_association_steps import \
    i_create_association, i_check_create_association
from bigmler.tests.basic_logistic_r_steps import \
    i_create_all_lr_resources, i_check_create_lr_model
from bigmler.tests.basic_deepnet_steps import \
    i_create_all_dn_resources, i_check_create_dn_model
from bigmler.tests.basic_pca_steps import \
    i_create_all_pca_resources, i_check_create_pca_model
from bigmler.tests.composite_steps import i_check_create_composite
from bigmler.tests.common_steps import shell_execute, check_debug, \
    base_shell_execute, check_rows_equal
from bigmler.reports import REPORTS_DIR

DECIMAL_PLACES = 3


def i_create_resources_from_ensemble_using_median(
    step, data=None, number_of_models=None, test=None, output=None):
    """Step: I create BigML resources from <data> using ensemble of
    <number_of_models> models to test <test> using median and log predictions
    in <output>
    """
    ok_(data is not None and test is not None and output is not None and
        number_of_models is not None)
    data = res_filename(data)
    test = res_filename(test)
    command = ("bigmler --train " + data + " --test " + test +
               " --store --output " + output +
               " --number-of-models " + number_of_models +
               " --median --max-batch-models 1 --no-fast")
    world.number_of_models = int(number_of_models)
    shell_execute(command, output, test=test)


def i_create_resources_in_prod_from_ensemble(
    step, data=None, number_of_models=None, test=None, output=None):
    """Step: I create BigML resources from <data> using ensemble of
    <number_of_models> models to test <test> and log predictions in <output>
    """
    i_create_resources_in_mode_from_ensemble(
        step, data=data, number_of_models=number_of_models,
        test=test, output=output)


def i_create_resources_in_mode_from_ensemble(
    step, data=None, number_of_models=None, test=None, output=None):
    """Step: I create BigML resources from <data> using ensemble of
    <number_of_models> models to test <test> and log predictions in
    <output>
    """
    ok_(data is not None and test is not None and output is not None and \
        number_of_models is not None)
    data = res_filename(data)
    test = res_filename(test)
    command = ("bigmler --train " + data + " --test " + test +
               " --store --output " + output +
               " --number-of-models " + number_of_models +
               " --max-batch-models 1 --no-fast")
    world.number_of_models = int(number_of_models)
    shell_execute(command, output, test=test)

def i_create_all_resources_with_split_field(step, data=None,
                                            split_field=None,
                                            objective=None,
                                            output_dir=None):
    """Step: I create BigML resources uploading train <data> file with
    split field <split_field> and log in <output_dir>
    """
    ok_(data is not None and split_field is not None
        and objective is not None and output_dir is not None)
    command = ("bigmler --train " + res_filename(data) +
               " --split-field \"" + split_field +
               "\" --objective \"" + objective +
               "\" --store --output-dir " + output_dir +
               " --max-batch-models 1 --no-fast")
    shell_execute(command, "%s/xxx" % output_dir)


def i_create_all_resources_with_focus_field(step, data=None,
                                            focus_field=None,
                                            objective=None,
                                            output_dir=None):
    """Step: I create BigML resources uploading train <data> file with
    focus field <focus_field> and log in <output_dir>
    """
    ok_(data is not None and focus_field is not None
        and objective is not None and output_dir is not None)
    command = ("bigmler --train " + res_filename(data) +
               " --focus-field \"" + focus_field +
               "\" --objective \"" + objective +
               "\" --store --output-dir " + output_dir +
               " --max-batch-models 1 --no-fast")
    shell_execute(command, "%s/xxx" % output_dir)


def i_create_all_resources_with_median(step, data=None,
                                       test=None, output=None):
    """Step: I create BigML resources uploading train <data> file using the
    median to test <test> and log predictions in <output>
    """
    ok_(data is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --store --output " + output +
               " --median --max-batch-models 1 --no-fast")
    shell_execute(command, output, test=test)


def i_create_all_resources_batch_to_dataset(step, data=None,
                                            test=None, output_dir=None):
    """Step: I create BigML resources uploading train <data> file to test
    <test> remotely to a dataset with no CSV output and log resources in
    <output_dir>
    """
    ok_(data is not None and test is not None and output_dir is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --store --no-csv --to-dataset --output-dir " +
               output_dir + " --remote")
    shell_execute(command, "%s/x.csv" % output_dir, test=test)


def i_create_source_with_locale(step, data=None, locale=None,
                                field_attributes=None, types=None,
                                output=None):
    """Step: I create a BigML source from file <data> with locale <locale>,
    field attributes <field_attributes> and types file <types> storing
    results in <output>
    """
    ok_(data is not None and locale is not None and output is not None and \
        types is not None and field_attributes is not None)
    field_attributes = res_filename(field_attributes)
    types = res_filename(types)
    command = ("bigmler --train " + res_filename(data) + " --locale " +
               locale + " --field-attributes " + field_attributes +
               " --types " + types + " --output " + output +
               " --no-dataset --no-model --store")
    shell_execute(command, output)


def i_create_composite_source(step, data=None, output=None):
    """Step: I create a BigML source from file <data> storing results in
    <output>
    """
    ok_(data is not None and output is not None)
    command = ("bigmler --train " + res_filename(data) + " --output " + output +
               " --no-dataset --no-model --store")
    shell_execute(command, output)


def i_create_source(step, data=None, output_dir=None):
    """Step: I create a BigML source from file <data> storing results in
    <output_dir>
    """
    ok_(data is not None and output_dir is not None)
    output = os.path.join(output_dir, "tmp.txt")
    command = ("bigmler source --data " + res_filename(data) +
               " --output-dir " + output_dir +
               " --store")
    shell_execute(command, output)


def i_create_all_resources_to_model(step, data=None, output=None):
    """Step: I create BigML resources uploading train <data> file to create
    model and log in <output>
    """
    ok_(data is not None and output is not None)
    command = ("bigmler --train " + res_filename(data) +
               " --store --output " + output + " --max-batch-models 1")
    shell_execute(command, output, test=None)


def i_create_kfold_cross_validation_objective(step, k_folds=None,
                                              objective=None, metric=None):
    """Step: I create BigML feature selection <k_folds>-fold cross-validations
    for <objective> improving <metric>
    """
    ok_(k_folds is not None and metric is not None and objective is not None)
    command = ("bigmler analyze --dataset " +
                             world.dataset['resource'] +
                             " --features --k-folds " + k_folds +
                             " --output " + world.directory +
                             " --optimize " + metric +
                             " --objective " + objective)
    shell_execute(command, os.path.join(world.directory, "tmp"))
    world.output = os.path.join(world.directory, "test", "kfold1",
                                "evaluation")


def i_create_kfold_cross_validation_options(step, k_folds=None, options=None):
    """Step: I create BigML feature selection <k_folds>-fold cross-validation
    with options <options>
    """
    ok_(k_folds is not None and options is not None)
    command = ("bigmler analyze --dataset " +
                             world.dataset['resource'] +
                             " --features --k-folds " + k_folds +
                             " --output " + world.directory +
                             options)
    shell_execute(command, os.path.join(world.directory, "tmp"))
    world.output = os.path.join(world.directory, "test", "kfold1",
                                "evaluation")

def i_create_all_resources_proportional(
    step, data=None, test=None, output=None):
    """Step: I create BigML resources uploading train <data> file to test
    <test> with proportional missing strategy and log predictions in
    <output>
    """
    ok_(data is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --missing-strategy proportional" +
               " --store --output " + output + " --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_create_all_resources_missing_splits(
    step, data=None, test=None, output=None):
    """Step: I create BigML resources uploading train <data> file to test
    <test> with a missing-splits model and log predictions in <output>
    """
    ok_(data is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --missing-splits" +
               " --store --output " + output + " --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_create_all_resources_remote_proportional(
    step, data=None, test=None, output=None):
    """Step: I create BigML resources uploading train <data> file to test
    <test> remotely with proportional missing strategy and log predictions
    in <output>
    """
    ok_(data is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --missing-strategy proportional --remote" +
               " --store --output " + output + " --max-batch-models 1")
    shell_execute(command, output, test=test, options="--no-header")


def i_create_all_resources_remote_missing_splits(
    step, data=None, test=None, output=None):
    """Step: I create BigML resources uploading train <data> file to test
    <test> remotely with a missing-splits model and log predictions in
    <output>
    """
    ok_(data is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --missing-splits --remote" +
               " --store --output " + output + " --max-batch-models 1")
    shell_execute(command, output, test=test, options="--no-header")


def i_create_all_resources_with_separator(
    step, data=None, test=None, output=None, separator=None):
    """Step: I create BigML resources uploading train <data> file to test
    <test> and log predictions in <output> with <separator>
    as test field separator
    """
    ok_(data is not None and test is not None and separator is not None and \
        output is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --test-separator " + separator + " --store --output " +
               output + " --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_create_all_resources_batch_map(
    step, data=None, test=None, fields_map=None, output=None):
    """Step: I create BigML resources uploading train <data> file to test
    <test> remotely with mapping file <fields_map> and log predictions in
    <output>
    """
    ok_(data is not None and test is not None and output is not None
        and fields_map is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --fields-map "
               + res_filename(fields_map) + " --store --output " + output +
               " --remote")
    shell_execute(command, output, test=test, options="--no-header")


def i_create_all_resources_batch(step, data=None, test=None, output=None):
    """Step: I create BigML resources uploading train <data> file to test
    <test> remotely and log predictions in <output>
    """
    ok_(data is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --store --output " + output + " --remote")
    shell_execute(command, output, test=test, options="--no-header")


def i_create_all_resources_with_no_headers(
    step, data=None, test=None, output=None):
    """Step: I create BigML resources uploading train <data> file with no
    headers to test <test> with no headers and log predictions in
    <output>
    """
    ok_(data is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --store --output " + output + " --max-batch-models 1 " +
               "--no-fast --no-train-header --no-test-header")
    shell_execute(command, output, test=test, options="--no-header")


def i_create_all_resources(step, data=None, test=None, output=None):
    """Step: I create BigML resources uploading train <data> file to test
    <test> and log predictions in <output>
    """
    ok_(data is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --store --output " + output +
               " --max-batch-models 1 --no-fast")
    shell_execute(command, output, test=test)


def i_create_all_resources_with_options(
    step, data=None, test=None, output=None, options=''):
    """Step: I create BigML resources uploading train <data> file to test
    <test> and log predictions in <output> with prediction options
    <options>
    """
    ok_(data is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --train " + res_filename(data) + " --test " + test +
               " --store --output " + output + " --max-batch-models 1 " +
               options.replace("'", "\""))
    shell_execute(command, output, test=test, options=options)


def i_create_resources_from_source_with_objective(
    step, multi_label=None, objective=None, model_fields=None,
    test=None, output=None):
    """Step: I create BigML <multi-label> resources using source with
    objective <objective> and model fields <model_fields> to test
    <test> and log predictions in <output>
    """
    ok_(test is not None and output is not None)
    test = res_filename(test)
    multi_label = "" if multi_label is None else " --multi-label "
    command = ("bigmler "+ multi_label +"--source " + world.source['resource']
               + " --objective " + objective + " --model-fields \" "
               + model_fields + "\" --test " + test
               + " --store --output " + output)
    shell_execute(command, output, test=test)


def i_create_resources_from_source(
    step, multi_label=None, test=None, output=None):
    """Step: I create BigML <multi-label> resources using source to test
    <test> and log predictions in <output>
    """
    ok_(test is not None and output is not None)
    test = res_filename(test)
    multi_label = "" if multi_label is None else " --multi-label "
    command = ("bigmler "+ multi_label +"--source " + world.source['resource']
               + " --test " + test + " --store --output " + output)
    shell_execute(command, output, test=test)


def i_create_resources_from_source_batch(step, output=None):
    """Step: I create BigML resources using source to test the previous test
    source remotely and log predictions in <output>
    """
    ok_(output is not None)
    command = ("bigmler --source " + world.source['resource']
               + " --test-source " + world.test_source['resource']
               + " --store --remote --output " + output)
    shell_execute(command, output)


def i_create_resources_from_dataset_with_objective(
    step, multi_label=None, objective=None, model_fields=None,
    test=None, output=None):
    """Step: I create BigML <multi-label> resources using dataset with
    objective <objective> and model fields <model_fields> to test <test>
    and log predictions in <output>
    """
    ok_(test is not None and output is not None)
    test = res_filename(test)
    multi_label = "" if multi_label is None else " --multi-label "
    command = ("bigmler "+ multi_label +"--dataset " +
               world.dataset['resource']
               + " --objective " + objective + " --model-fields \" "
               + model_fields + "\" --test " + test
               + " --store --output " + output)
    shell_execute(command, output, test=test)


def i_create_resources_from_dataset(
    step, multi_label=None, test=None, output=None):
    """Step: I create BigML <multi-label> resources using dataset to test
    <test> and log predictions in <output>
    """
    ok_(test is not None and output is not None)
    multi_label = "" if multi_label is None else " --multi-label "
    test = res_filename(test)
    command = ("bigmler "+ multi_label +"--dataset " +
               world.dataset['resource'] + " --test " + test +
               " --store --output " + output)
    shell_execute(command, output, test=test)


def i_create_resources_from_model_batch(step, fields=None, output=None):
    """Step: I create BigML resources using a model to test the previous test
    dataset remotely with prediction headers and fields <fields> and log
    predictions in <output>
    """
    ok_(output is not None and fields is not None)
    command = ("bigmler --model " + world.model['resource'] +
               " --test-dataset " +
               world.test_dataset['resource'] + " --store --remote " +
               "--prediction-header --prediction-info full " +
               "--prediction-fields \"" + fields + "\" --output " + output)
    shell_execute(command, output,
                  test_rows=world.test_dataset['object']['rows'] + 1)


def i_create_resources_from_dataset_batch(step, output=None):
    """Step: I create BigML resources using dataset to test the previous test
    dataset remotely and log predictions in <output>
    """
    ok_(output is not None)
    command = ("bigmler --dataset " + world.dataset['resource'] +
               " --test-dataset " +
               world.test_dataset['resource'] + " --store --remote --output "
               + output)
    shell_execute(command, output,
                  test_rows=world.test_dataset['object']['rows'])


def i_create_resources_from_dataset_objective_model(
    step, objective=None, fields=None, test=None, output=None):
    """Step: I create BigML resources using dataset, objective field
    <objective> and model fields <model_fields> to test <test> and log
    predictions in <output>
    """
    ok_(objective is not None and fields is not None and test is not None
        and output is not None)
    test = res_filename(test)
    command = ("bigmler --dataset " + world.dataset['resource'] +
               " --objective " + objective + " --model-fields " +
               fields + " --test " + test + " --store --output " + output)
    shell_execute(command, output, test=test)


def i_create_resources_from_local_model(
    step, directory=None, test=None, output=None):
    """Step: I create BigML resources using local model in <directory> to test
    <test> and log predictions in <output>
    """
    ok_(test is not None and output is not None and directory is not None)
    test = res_filename(test)
    with open(os.path.join(directory, "models")) as model_file:
        model_id = model_file.read().strip()
    command = ("bigmler --model-file " +
               storage_file_name(directory, model_id) +
               " --test " +
               test + " --store --output " + output + " --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_create_resources_from_model(step, test=None, output=None):
    """Step: I create BigML resources using model to test <test> and
    log predictions in <output>
    """
    ok_(test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --model " + world.model['resource'] + " --test " +
               test + " --store --output " + output + " --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_create_resources_from_model_with_op(step, operating_point=None,
                                          test=None, output=None):
    """Step: I create BigML resources using model with operating point
    <operating_point> to test <test> and log predictions in <output>
    """
    ok_(operating_point is not None and
        test is not None and output is not None)
    test = res_filename(test)
    operating_point = res_filename(operating_point)
    command = ("bigmler --model " + world.model['resource'] + " --test " +
               test + " --operating-point " + operating_point +
               " --store --output " + output + " --max-batch-models 1")
    shell_execute(command, output, test=test)

def i_create_resources_from_ensemble_with_threshold(
    step, test=None, output2=None, output3=None):
    """Step: I create BigML resources using the previous ensemble with
    different thresholds to test <test> and log predictions in <output2> and
    <output3>
    """
    ok_(test is not None and output2 is not None and output3 is not None)
    message = None
    try:
        test = res_filename(test)
        command = ("bigmler --ensemble " + world.ensemble['resource'] +
                   " --test " + test + " --tag my_ensemble --store --output " +
                   output2 + " --method threshold --threshold " +
                   str(world.number_of_models))
        command = check_debug(command)
        retcode = check_call(command, shell=True)
        ok_(retcode >= 0)
        command = ("bigmler --ensemble " + world.ensemble['resource'] +
                   " --test " + test + " --tag my_ensemble --store --output " +
                   output3 + " --method threshold --threshold 1")
        command = check_debug(command)
        retcode = check_call(command, shell=True)
        ok_(retcode >= 0)
    except (OSError, CalledProcessError, IOError) as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_create_resources_from_ensemble_with_threshold_rem(
    step, test=None, output2=None, output3=None, threshold_class=None):
    """Step: I create BigML resources using the previous ensemble with
    different thresholds to test <test> remotely and log predictions in
    <output2> and <output3>
    """
    ok_(test is not None and output2 is not None and output3 is not None
        and threshold_class is not None)
    message = None
    try:
        test = res_filename(test)
        command = ("bigmler --ensemble " + world.ensemble['resource'] +
                   " --test " + test + " --tag my_ensemble --store --output " +
                   output2 + ' --class ' + threshold_class +
                   " --remote --method threshold" +
                   " --threshold " +
                   str(world.number_of_models))
        command = check_debug(command)
        retcode = check_call(command, shell=True)
        ok_(retcode >= 0)
        command = ("bigmler --ensemble " + world.ensemble['resource'] +
                   " --test " + test + " --tag my_ensemble --store --output " +
                   output3 + ' --class ' + threshold_class +
                   " --remote --method threshold" +
                   " --threshold 1")
        command = check_debug(command)
        retcode = check_call(command, shell=True)
        ok_(retcode >= 0)
    except (OSError, CalledProcessError, IOError) as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_create_resources_from_local_ensemble(step, number_of_models=None,
                                           directory=None, test=None,
                                           output=None):
    """Step: I create BigML resources using local ensemble of <number_of_models>
    models in <directory> to test <test> and log predictions in <output>
    """
    ok_(number_of_models is not None and test is not None and
        output is not None and directory is not None)
    world.directory = os.path.dirname(output)
    world.folders.append(world.directory)
    message = None
    with open(os.path.join(directory, "ensembles")) as ensemble_file:
        ensemble_id = ensemble_file.read().strip()
    try:
        test = res_filename(test)
        command = ("bigmler --ensemble-file " +
                   storage_file_name(directory, ensemble_id) +
                   " --test " + test + " --store" +
                   " --output " + output)
        command = check_debug(command)
        retcode = check_call(command, shell=True)
        ok_(retcode >= 0)

        world.test_lines = file_number_of_lines(test)
        # test file has headers in it, so first line must be ignored
        world.test_lines -= 1
        world.output = output
        world.number_of_models = len(world.ensemble['object']['models'])
    except (OSError, CalledProcessError, IOError) as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_create_resources_from_local_ensemble_with_op(
        step, number_of_models=None,
        directory=None, test=None,
        output=None, operating_point=None):
    """Step: I create BigML resources using local ensemble of
    <number_of_models> models with operating point <operating_point> in
    <idrectory> to test <test> and log predictions in <output>
    """
    ok_(number_of_models is not None and test is not None and
        output is not None and directory is not None and
        operating_point is not None)
    world.directory = os.path.dirname(output)
    world.folders.append(world.directory)
    message = None
    with open(os.path.join(directory, "ensembles")) as ensemble_file:
        ensemble_id = ensemble_file.read().strip()
    try:
        test = res_filename(test)
        operating_point = res_filename(operating_point)
        command = ("bigmler --ensemble-file " +
                   storage_file_name(directory, ensemble_id) +
                   " --test " + test + " --store" +
                   " --output " + output + " --operating-point " +
                   operating_point)
        command = check_debug(command)
        retcode = check_call(command, shell=True)
        ok_(retcode >= 0)

        world.test_lines = file_number_of_lines(test)
        # test file has headers in it, so first line must be ignored
        world.test_lines -= 1
        world.output = output
        world.number_of_models = len(world.ensemble['object']['models'])
    except (OSError, CalledProcessError, IOError) as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_create_resources_from_ensemble_with_replacement(
    step, number_of_models=None, test=None, output=None):
    """Step: I create BigML resources using ensemble of <number_of_models>
    models with replacement to test <test> and log predictions in
    <output>
    """
    i_create_resources_from_ensemble_generic(step, number_of_models, "",
                                             test, output)


def i_create_resources_from_ensemble(
    step, number_of_models=None, test=None, output=None):
    """Step: I create BigML resources using ensemble of <number_of_models>
    models to test <test> and log predictions in <output>
    """
    i_create_resources_from_ensemble_generic(step, number_of_models,
        " --no-fast --ensemble-sample-no-replacement",
        test, output)


def i_create_resources_from_boosted_ensemble(
    step, iterations=None, test=None, output=None):
    """Step: I create BigML resources using boosted ensemble in
    <iterations> iterations to test <test> and log predictions in <output>
    """
    ok_(iterations is not None and test is not None and
        output is not None)
    test = res_filename(test)
    command = ("bigmler --dataset " + world.dataset['resource'] +
               " --test " + test + " --boosting-iterations " +
               str(iterations) + " --tag my_ensemble --store" +
               " --output " + output)
    shell_execute(command, output, test=test)


def i_create_resources_remotely_from_boosted_ensemble(
    step, iterations=None, test=None, output=None):
    """Step: I create BigML resources using boosted ensemble in <iterations>
    iterations to remotely test <test> and log predictions in <output>
    """
    ok_(iterations is not None and test is not None and \
        output is not None)
    test = res_filename(test)
    command = ("bigmler --dataset " + world.dataset['resource'] +
               " --test " + test + " --boosting-iterations " +
               str(iterations) + " --remote --tag my_ensemble --store" +
               " --output " + output + " --to-csv")
    shell_execute(command, output, test=test, options="--no-header")


def i_create_resources_from_ensemble_generic(step, number_of_models=None,
    no_replacement="", test=None, output=None):
    """Creating bagging ensemble with or withour replacement"""
    ok_(number_of_models is not None and test is not None and \
        output is not None)
    test = res_filename(test)
    command = ("bigmler --dataset " + world.dataset['resource'] +
               " --test " + test + " --number-of-models " +
               str(number_of_models) + " --tag my_ensemble --store" +
               " --output " + output + no_replacement)
    shell_execute(command, output, test=test)


def i_create_resources_from_models_file_with_objective(
    step, multi_label=None, models_file=None, objective=None,
    test=None, output=None):
    """Step: I create BigML <multi-label> resources using models in file
    <models_file> with objective <objective> to test <test> and log
    predictions in <output>
    """
    ok_(models_file is not None and test is not None and output is not None
        and objective is not None)
    test = res_filename(test)
    multi_label = "" if multi_label is None else " --multi-label "
    command = ("bigmler "+ multi_label +"--models " + models_file + " --test "
               + test + " --store --output " + output
               + " --objective " + objective)
    shell_execute(command, output, test=test)


def i_create_resources_from_models_file(
    step, multi_label=None, models_file=None, test=None, output=None):
    """Step: I create BigML <multi-label> resources using models in file
    <models_file> to test <test> and log predictions in <output>
    """
    ok_(models_file is not None and test is not None and output is not None)
    test = res_filename(test)
    multi_label = "" if multi_label is None else " --multi-label "
    command = ("bigmler "+ multi_label +"--models " + models_file + " --test "
               + test + " --store --output " + output)
    shell_execute(command, output, test=test)


def i_create_resources_from_dataset_file(
    step, dataset_file=None, test=None, output=None):
    """Step: I create BigML resources using dataset in file <dataset_file>
    to test <test> and log predictions in <output>
    """
    ok_(dataset_file is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --datasets " + dataset_file + " --test " + test +
               " --store --output " + output)
    shell_execute(command, output, test=test)


def i_create_cross_validation_from_dataset(
    step, rate=None, dataset_file=None, output=None):
    """Step: I create a BigML cross-validation with rate <rate> using the
    # dataset in file <dateset_file> and log results in <output>
    """
    ok_(rate is not None and output is not None and dataset_file is not None)
    with open(dataset_file, "r") as handler:
        dataset_id = handler.readline().strip()
    world.directory = os.path.dirname(output)
    world.folders.append(world.directory)
    world.number_of_models = int(MONTECARLO_FACTOR * float(rate))
    world.number_of_evaluations = world.number_of_models
    message = None
    try:
        command = ("bigmler --dataset " + dataset_id +
                   " --cross-validation-rate " + rate + " --store --output "
                   + output)
        command = check_debug(command)
        retcode = check_call(command, shell=True)
        ok_(retcode >= 0)
        world.output = output
    except (OSError, CalledProcessError, IOError) as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_find_predictions_files(
    step, directory1=None, directory2=None, output=None):
    """Step: I combine BigML predictions files in <directory1> and <directory2>
    into <output>
    """
    ok_(directory1 is not None and directory2 is not None and
        output is not None)
    world.directory = os.path.dirname(output)
    world.folders.append(world.directory)
    message = None
    try:
        command = ("bigmler --combine-votes " + directory1 + "," + directory2 +
                   " --store --output " + output)
        command = check_debug(command)
        retcode = check_call(command, shell=True)
        ok_(retcode >= 0)
        world.test_lines = file_number_of_lines(
            os.path.join(directory1, "predictions.csv"))
        world.output = output
    except (OSError, CalledProcessError, IOError) as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_find_predictions_files_with_method(
    step, directory1=None, directory2=None, output=None, method=None):
    """Step: I combine BigML predictions files in <directory1> and
    <directory2> into <output> with method <method>
    """
    ok_(directory1 is not None and directory2 is not None and
        output is not None and method is not None)
    world.directory = os.path.dirname(output)
    world.folders.append(world.directory)
    message = None
    try:
        command = ("bigmler --combine-votes " + directory1 + "," + directory2 +
                   " --store --output " + output + " --method " + method)
        command = check_debug(command)
        retcode = check_call(command, shell=True)
        ok_(retcode >= 0)
        world.test_lines = file_number_of_lines(
            os.path.join(directory1, "predictions.csv"))
        world.output = output
    except (OSError, CalledProcessError, IOError) as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_create_balanced_model(step, data=None, output_dir=None):
    """Step: I create a BigML balanced model from <data> and store logs in
    <output_dir>
    """
    ok_(data is not None and output_dir is not None)
    if not data.startswith("https"):
        data = res_filename(data)
    command = ("bigmler --train " + data + " --balance " +
               " --store --output-dir " + output_dir)
    base_shell_execute(command, output_dir)


def i_create_balanced_model_from_sample(step, data=None, output_dir=None):
    """Step: I create a BigML balanced model from <data> sampling 50% and store
    logs in <output_dir>
    """
    ok_(data is not None and output_dir is not None)
    if not data.startswith("https"):
        data = res_filename(data)
    command = ("bigmler --train " + data +
               " --store --no-model --output-dir " + output_dir)
    base_shell_execute(command, output_dir)
    command = ("bigmler --datasets " + output_dir + "/dataset" +
               " --store --to-dataset --sample-rate 0.5 --no-model "+
               " --output-dir " +
               output_dir)
    base_shell_execute(command, output_dir)

    command = ("bigmler --datasets " + output_dir + "/dataset_gen" +
               " --store --balance  --output-dir " +
               output_dir)
    base_shell_execute(command, output_dir)


def i_create_weighted_field_model(
    step, data=None, field=None, output_dir=None, objective=None):
    """Step: I create a BigML field weighted model from <data> using field
    <field> as weight and store logs in <objective>
    """
    ok_(data is not None and field is not None and output_dir is not None and
        objective is not None)
    command = ("bigmler --train " + res_filename(data) +
               " --weight-field " + field + " --objective " + objective +
               " --store --output-dir " + output_dir)
    base_shell_execute(command, output_dir)


def i_create_objective_weighted_model(
    step, data=None, path=None, output_dir=None):
    """Step: I create a BigML objective weighted model from <data> using the
    objective weights in file <path> and store logs in <output_dir>
    """
    ok_(data is not None and path is not None and output_dir is not None)
    command = ("bigmler --train " + res_filename(data) +
               " --objective-weights " + res_filename(path) +
               " --store --output-dir " + output_dir)
    base_shell_execute(command, output_dir)


def i_retrain_model(step, data=None, output_dir=None):
    """Step: I retrain the model from <data> and store logs in <output>"""
    ok_(data is not None and output_dir is not None)
    world.origin_model = world.model
    if not data.startswith("https"):
        data = res_filename(data)
    command = ("bigmler retrain --add " + data +
               " --id " + world.model['resource'] +
               " --store --output-dir " + output_dir)
    base_shell_execute(command, output_dir)


def i_create_source_from_connector(step, data=None, output_dir=None,
                                   query=None):
    """Step: I create source from external connector"""
    ok_(data is not None and output_dir is not None and query is not None)
    connector_id = world.external_connector["resource"].replace(
        "externalconnector/", "")
    with open(data, 'w+') as file_handler:
        json.dump({"source": "postgresql",
                   "externalconnector_id": connector_id,
                   "query": query}, file_handler)
    command = ("bigmler --train " + os.path.abspath(data) +
               " --store --no-dataset --output-dir " + output_dir)
    base_shell_execute(command, output_dir)


def i_check_create_source(step):
    """Step: I check that the source has been created"""
    source_file = os.path.join(world.directory, "source")
    message = None
    try:
        with open(source_file) as handler:
            source = check_resource(
                handler.readline().strip(), world.api.get_source)
            if source['resource'] not in world.sources:
                world.sources.append(source['resource'])
            world.source = source
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_model_double(step):
    """Step: I check that the model has doubled its rows"""
    ok_(world.model['object']['rows'] == \
        2 * world.origin_model['object']['rows'])


def i_check_create_dataset(step, suffix=None):
    """Step: I check that the <test|train> dataset has been created"""
    suffix = "" if suffix is None else "_%s" % suffix[:-1]
    dataset_file = os.path.join(world.directory, f"dataset{suffix}")
    message = None
    try:
        with open(dataset_file) as handler:
            for dataset_id in handler:
                dataset_id = dataset_id.strip()
                dataset = check_resource(dataset_id,
                                         world.api.get_dataset)
                assert (not 'user_metadata' in dataset['object'] or
                        not 'max_categories'
                        in dataset['object']['user_metadata'])
                if dataset['resource'] not in world.datasets:
                    world.datasets.append(dataset['resource'])
                world.dataset = dataset
    except Exception:
        message = traceback.format_exc()
    ok_(message is None, msg=message)


def i_check_create_new_dataset(step):
    """Step: I check that the new dataset has been created"""
    dataset_file = os.path.join(world.directory, "dataset_gen")
    message = None
    try:
        with open(dataset_file) as handler:
            dataset_id = handler.readline().strip()
            dataset = check_resource(dataset_id,
                                     world.api.get_dataset)
            world.datasets.append(dataset['resource'])
            world.dataset = dataset
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_create_dataset_shared(step):
    """Step: I check that the dataset has been created and shared"""
    i_check_create_dataset(step)
    assert world.dataset['object']['shared']


def i_check_create_model(step):
    """Step: I check that the model has been created"""
    model_file = os.path.join(world.directory, "models")
    message = None
    try:
        with open(model_file) as handler:
            model = check_resource(handler.readline().strip(),
                                   world.api.get_model)
            world.models.append(model['resource'])
            world.model = model
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_create_model_in_execution(step):
    """Step: I check that the model has been created in the execution"""
    output_resources = world.execution[
        "object"]["execution"]["output_resources"]
    for resource in output_resources:
        try:
            model = check_resource(resource["id"],
                                   world.api.get_model)
            if model is not None:
                world.models.append(model['resource'])
                world.model = model
                break
        except Exception:
            continue


def i_check_create_kfold_models(step, kfolds):
    """Step: I check that the <kfolds>-models have been created"""
    directory = os.path.dirname(os.path.dirname(world.output))
    directories = [os.path.join(directory, folder)
                   for folder in os.listdir(directory) if
                   os.path.isdir(os.path.join(directory, folder))]

    for directory in directories:
        if not directory.endswith("_pred"):
            model_file = os.path.join(directory, "models")
            try:
                with open(model_file) as models_file:
                    models_list = list(map(str.strip, models_file.readlines()))
                world.models.extend(models_list)
                world.model = models_list[-1]
                eq_(int(kfolds), len(models_list))
            except Exception as exc:
                ok_(False, msg=str(exc))


def i_check_create_kfold_ensembles(step, kfolds):
    """Step: I check that the <kfolds>-ensembles have been created"""
    directory = os.path.dirname(os.path.dirname(world.output))
    directories = [os.path.join(directory, folder)
                   for folder in os.listdir(directory) if
                   os.path.isdir(os.path.join(directory, folder))]
    for directory in directories:
        if not directory.endswith("_pred"):
            model_file = os.path.join(directory, "ensembles")
            try:
                with open(model_file) as models_file:
                    models_list = list(map(str.strip, models_file.readlines()))
                world.ensembles.extend(models_list)
                world.ensemble = models_list[-1]
                eq_(int(kfolds), len(models_list))
            except Exception as exc:
                ok_(False, msg=str(exc))


def i_check_create_model_shared(step):
    """Step: I check that the model has been created and shared"""
    i_check_create_model(step)
    ok_(world.model['object']['shared'])


def i_check_model_is_balanced(step):
    """Step: I check that the model is balanced"""
    assert ('balance_objective' in world.model['object'] and
            world.model['object']['balance_objective'])


def i_check_weighted_model(step, field=None):
    """Step: I check that the model uses as weight <field>"""
    ok_(field is not None, msg="No field provided")
    ok_('weight_field' in world.model['object'] and
        world.model['object']['weight_field'] == field)


def i_check_first_node_children(step, children=None, field=None):
    """Step: I check that the first node has <children> branches"""
    ok_(children is not None and field is not None)
    root_children = world.model['object'] and \
        len(world.model['object']['model']['root']['children'])
    root_field = world.model['object']['model']['root'][
        'children'][0]['predicate']['field']
    eq_(root_children, children)
    eq_(root_field, field)


def i_check_objective_weighted_model(step, weights=None):
    """Step: check that the model uses as objective weights <weights>"""
    ok_(weights is not None)
    ok_('objective_weights' in world.model['object'] and
        world.model['object']['objective_weights'] == json.loads(weights))


def i_check_create_ensemble(step):
    """Step: I check that the ensemble has been created"""
    ensemble_file = os.path.join(world.directory, "ensembles")
    message = None
    try:
        with open(ensemble_file) as handler:
            ensemble = check_resource(handler.readline().strip(),
                                      world.api.get_ensemble)
            world.ensembles.append(ensemble['resource'])
            world.ensemble = ensemble
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_create_models(step):
    """Step: I check that the models have been created"""
    i_check_create_models_in_ensembles(step, False)


def i_check_create_models_in_ensembles(step, in_ensemble=True):
    """Step: I check that the models in the ensembles have been created"""
    model_file = os.path.join(world.directory, "models")
    number_of_lines = 0
    count = 0
    message = None
    while world.number_of_models != number_of_lines and count < 10:
        number_of_lines = 0
        model_ids = []
        with open(model_file) as handler:
            for line in handler:
                number_of_lines += 1
                model_id = line.strip()
                model_ids.append(model_id)
            if world.number_of_models != number_of_lines:
                time.sleep(10)
                count += 1
    eq_(world.number_of_models, number_of_lines,
        msg="number of models %s and number of lines in models file %s: %s" % \
            (world.number_of_models, model_file, number_of_lines))
    world.model_ids = model_ids
    for model_id in model_ids:
        try:
            model = check_resource(model_id, world.api.get_model)
            if in_ensemble:
                ensemble_id = "ensemble/%s" % model['object']['ensemble_id']
                if not ensemble_id in world.ensembles:
                    world.ensembles.append(ensemble_id)
            else:
                world.models.append(model_id)
        except Exception as exc:
            message = str(exc)
            break
    ok_(message is None, msg=message)


def i_check_create_evaluation(step):
    """Step: I check that the evaluation has been created"""
    evaluation_file = os.path.join(world.directory, "evaluations")
    message = None
    try:
        with open(evaluation_file) as handler:
            evaluation = check_resource(handler.readline().strip(),
                                        world.api.get_evaluation)
            world.evaluations.append(evaluation['resource'])
            world.evaluation = evaluation
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def check_create_kfold_cross_validation(step, kfolds, directory):
    """Checking that all evaluations in a cross-validation are created"""
    evaluations_file = os.path.join(directory, "evaluations")
    message = None
    try:
        with open(evaluations_file, "r") as evaluations_file:
            evaluations_list = list(
                map(str.strip, evaluations_file.readlines()))
        world.evaluations.extend(evaluations_list)
        world.evaluation = evaluations_list[-1]
        eq_(int(kfolds), len(evaluations_list))
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_create_all_kfold_cross_validations(step, kfolds):
    """Step: I check that all the kfolds-fold cross-validations have been
    created
    """
    directory = os.path.dirname(os.path.dirname(world.output))
    directories = [os.path.join(directory, folder)
                   for folder in os.listdir(directory) if
                   os.path.isdir(os.path.join(directory, folder))]
    for directory in directories:
        if not directory.endswith("_pred"):
            check_create_kfold_cross_validation(step, kfolds, directory)


def i_check_create_kfold_cross_validation(step, kfolds):
    """Step: I check that the <kfolds>-fold cross-validation has been created
    """
    check_create_kfold_cross_validation(step, kfolds,
                                        os.path.dirname(world.output))


def i_check_create_kfold_datasets(step, kfolds):
    """Step: I check that the <kfolds>-datasets have been created"""
    datasets_file = os.path.join(os.path.dirname(
        os.path.dirname(world.output)), "dataset_gen")
    message = None
    try:
        with open(datasets_file) as handler:
            datasets_list = list(map(str.strip, handler.readlines()))
        world.datasets.extend(datasets_list)
        world.dataset = datasets_list[-1]
        eq_(int(kfolds), len(datasets_list))
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_feature_selection(step, selection, metric, metric_value):
    """Step: the best feature selection is <selection>, with <metric> of
    <metric_value>
    """
    ok_(selection is not None and metric is not None and \
        metric_value is not None)
    sessions_file = os.path.join(world.directory, "bigmler_sessions")
    message = None
    try:
        with open(sessions_file, open_mode("r")) as sessions_file:
            content = sessions_file.read()
        text = "The best feature subset is: %s \n%s = %s" % (
            selection, metric.capitalize(), metric_value)
        ok_(content.find(text) > -1)
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_node_threshold(step, node_threshold, metric, metric_value):
    """Step: the best node threshold is <node_threshold>, with <metric>
    of <metric_value>
    """
    ok_(node_threshold is not None and metric is not None and \
        metric_value is not None)
    sessions_file = os.path.join(world.directory, "bigmler_sessions")
    message = None
    try:
        with open(sessions_file, open_mode("r")) as handler:
            content = handler.read()
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)
    text = "The best node threshold is: %s \n%s = %s" % (
        node_threshold, metric.capitalize(), metric_value)
    ok_(content.find(text) > -1)


def i_check_create_evaluation_shared(step):
    """Step: I check that the evaluation has been created and shared"""
    i_check_create_evaluation(step)
    ok_(world.evaluation['object']['shared'])


def i_check_gazibit_reports(step, shared=''):
    """Step I check that the Gazibit (shared )?report has been created"""
    shared = "_%s" % shared[0: -1] if shared is not None else ''
    gazibit_file = os.path.join(world.directory,"reports", "gazibit%s.json" %
        shared)
    message = None
    try:
        with open(gazibit_file) as handler:
            content = handler.read()
        ok_(content.find('%START_BIGML_') < 0 and
            content.find('%END_BIGML_') < 0 and
            content.find('%BIGML_') < 0)
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_create_evaluations(step, number_of_evaluations=None):
    """Step: I check that the <number_of_evaluations> evaluations
    have been created
    """
    if number_of_evaluations is not None:
        world.number_of_evaluations = int(number_of_evaluations)
    evaluations_file = os.path.join(world.directory, "evaluations")
    evaluation_ids = []
    number_of_lines = 0
    count = 0
    message = None
    while world.number_of_evaluations != number_of_lines and count < 10:
        number_of_lines = 0
        with open(evaluations_file) as handler:
            for line in handler:
                number_of_lines += 1
                evaluation_id = line.strip()
                evaluation_ids.append(evaluation_id)
        if world.number_of_evaluations != number_of_lines:
            time.sleep(10)
            count += 1
    eq_(world.number_of_evaluations, number_of_lines,
        msg="number of evaluations %s and number of lines in"
            " evaluations file %s: %s" % (
            world.number_of_evaluations, evaluations_file, number_of_lines))
    world.evaluation_ids = evaluation_ids
    for evaluation_id in evaluation_ids:
        try:
            check_resource(evaluation_id, world.api.get_evaluation)
            world.evaluations.append(evaluation_id)
        except Exception as exc:
            message = str(exc)
            break
    ok_(message is None, msg=message)


def i_check_create_predictions(step):
    """Step: I check that the predictions are ready"""
    predictions_lines = file_number_of_lines(world.output)
    eq_(predictions_lines, world.test_lines)


def i_check_create_projections(step):
    """Step: I check that the projections are ready"""
    i_check_create_predictions(step)


def i_check_predictions_file(step, predictions_file, check_file):
    """Step: the predictions file <predictions_file> is like <check_file>"""
    world.output = predictions_file
    i_check_predictions(step, check_file)


def i_check_predictions(step, check_file, headers=False):
    """Step: the local prediction file is like <check_file>"""
    check_file_path = res_filename(check_file)
    predictions_file_path = world.output
    message = None
    try:
        message = check_rows_equal(predictions_file_path,
            check_file_path)
        if message is not None:
            shutil.copyfile(predictions_file_path, "%s.new" % check_file_path)
    except Exception:
        shutil.copyfile(predictions_file_path, "%s.new" % check_file_path)
        message = traceback.format_exc()
    ok_(message is None, msg=message)


def i_check_projections(step, check_file):
    """Checking that projections have been created"""
    i_check_predictions(step, check_file)


def i_check_predictions_with_different_thresholds(step, output2, output3):
    """Step: local predictions for different thresholds in <output2>
    and <output3> are different
    """
    try:
        with open(output2) as handler2:
            predictions_file = handler2.read()
        with open(output3) as handler3:
            predictions_file_k = handler3.read()
        ok_(predictions_file != predictions_file_k)
    except Exception as exc:
        ok_(False, msg=str(exc))


def i_check_cross_validation(step, check_file):
    """Step: the cross-validation json model info is like the one in
    <check_file>
    """
    check_file = res_filename(check_file)
    cv_file = os.path.join(world.directory, "cross_validation.json")
    try:
        with open(cv_file) as cv_handler:
            cv_content = json.loads(cv_handler.read())
        with open(check_file) as check_handler:
            check_content = json.loads(check_handler.read())
    except Exception as exc:
        ok_(False, msg=str(exc))
    eq_(cv_content['model']['confusion_matrix'],
        check_content['model']['confusion_matrix'])


def i_check_stored_source(step):
    """Step: I check that the stored source file exists"""
    source_file = os.path.join(world.directory, "source")
    message = None
    try:
        with open(source_file) as handler:
            source_id = handler.readline().strip()
            world.sources.append(source_id)
        storage_source_file = os.path.join(
            world.directory, source_id.replace("/", "_"))
        if os.path.exists(storage_source_file):
            with open(storage_source_file) as handler:
                world.source = json.loads(handler.read().strip())
        else:
            ok_(False)
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_source_locale(step, bigml_locale):
    """Step: And the locale of the source is <bigml_locale>"""
    try:
        locale = world.source['object']["source_parser"]["locale"]
        eq_(bigml_locale, locale)
    except KeyError as exc:
        ok_(False, msg=str(exc))


def i_check_source_type(step, field_id, field_type):
    """Step: the type of field <field_id> is <field_type>"""
    try:
        source_type = world.source['object']["fields"][field_id]['optype']
        eq_(source_type, field_type)
    except KeyError as exc:
        ok_(False, msg=str(exc))


def i_check_source_label(step, field_id, field_label):
    """Step: the label of field <field_id> is <field_label>"""
    try:
        label = world.source['object']["fields"][field_id]['label']
        eq_(label, field_label)
    except KeyError as exc:
        ok_(False, msg=str(exc))

def check_source_exists(step, exists=True):
    """Step: I check that the source exists"""
    source = world.api.get_source(step.source)
    eq_(world.api.ok(source), exists)


def i_create_all_resources_to_evaluate(step, data=None, output=None):
    """Step: I create BigML resources uploading train <data> file to evaluate
    and log evaluation in <output>
    """
    ok_(data is not None and output is not None)
    command = "bigmler --evaluate --train " + res_filename(data) + \
              " --store --output " + output
    shell_execute(command, output)


def i_create_all_resources_to_evaluate_and_report(
    step, data=None, output=None):
    """Step: I create BigML resources and share them uploading train <data>
    file to evaluate and log evaluation and reports in <output>
    """
    ok_(data is not None and output is not None)
    command = ("bigmler --evaluate --shared --report gazibit" +
               " --train " + res_filename(data) +
               " --store --no-upload --output " + output)
    shell_execute(command, output)

def i_create_all_resources_to_evaluate_and_share(step, data=None, output=None):
    """Step: I create BigML resources and share them uploading train <data>
    file to evaluate and log evaluation in <output>
    """
    ok_(data is not None and output is not None)
    command = ("bigmler --evaluate --shared --train " + res_filename(data) +
               " --store --output " + output)
    command = check_debug(command)
    shell_execute(command, output)

def i_create_dataset_with_attributes(
    step, data=None, attributes=None, output=None):
    """Step: I create BigML dataset uploading train <data> file with attributes
    <attributes> in <output>
    """
    ok_(data is not None and output is not None and attributes is not None)
    command = ("bigmler --train " + res_filename(data) +
               " --source-attributes " + res_filename(attributes) +
               " --no-model --store --output " + output)
    shell_execute(command, output)

def i_create_dataset(step, data=None, output=None):
    """Step: I create BigML dataset uploading train <data> file in <output>"""
    ok_(data is not None and output is not None)
    command = ("bigmler --train " + res_filename(data) +
               " --no-model --store --output " + output)
    shell_execute(command, output)


def i_create_kfold_cross_validation(step, k_folds=None):
    """Step: I create BigML <k_folds>-fold cross-validation"""
    ok_(k_folds is not None)
    output = os.path.abspath(
        os.path.join(world.directory, "test", "k_fold0", "evaluation"))
    command = ("bigmler analyze --dataset " +
               world.dataset['resource'] +
               " --cross-validation --k-folds " + k_folds +
               " --output " + world.directory)
    base_shell_execute(command, os.path.join(world.directory, "tmp"))
    world.output = output

def i_create_nodes_analysis(step, min_nodes=None, max_nodes=None,
                            nodes_step=None, k_fold=None, metric=None):
    """Step: I create BigML nodes analysis from <min_nodes> to
    <max_nodes> by <nodes_step> with <k_fold>-cross-validation
    improving <metric>
    """
    ok_(max_nodes is not None and \
        nodes_step is not None and k_fold is not None and \
        metric is not None)
    output = os.path.abspath(
        os.path.join(world.directory, "test", "k_fold0", "evaluation"))
    command = ("bigmler analyze --dataset " + world.dataset['resource'] +
               " --nodes --min-nodes " + min_nodes +
               " --max-nodes " + max_nodes +
               " --nodes-step " + nodes_step +
               " --k-folds " + k_fold +
               " --output " + world.directory +
               " --optimize " + metric)
    base_shell_execute(command, world.directory)
    world.output = output


def i_create_nodes_analysis_from_dataset_file(
    step, min_nodes=None, max_nodes=None,
    nodes_step=None, k_fold=None, metric=None):
    """Step: I create BigML nodes analysis from datasets file from <min_nodes>
    to <max_nodes> by <nodes_step> with <k_fold>-cross-validation improving
    <metric>
    """
    ok_(max_nodes is not None and \
        nodes_step is not None and k_fold is not None and \
        metric is not None)
    output = os.path.abspath(
        os.path.join(world.directory, "test", "k_fold1", "evaluation"))
    command = ("bigmler analyze --datasets %s/dataset" % world.directory +
               " --nodes --min-nodes " + min_nodes +
               " --max-nodes " + max_nodes +
               " --nodes-step " + nodes_step +
               " --k-folds " + k_fold +
               " --output " + world.directory +
               " --optimize " + metric)
    base_shell_execute(command, world.directory)
    world.output = output


def i_create_kfold_cross_validation_separator_metric_no_fields(
    step, k_folds=None, features=None, args_separator=None, metric=None):
    """Step: I create BigML feature selection <k_folds>-fold cross-validations
    excluding <features> with separator <args_separator> improving
    <metric>
    """
    ok_(k_folds is not None and metric is not None and features is not None
        and args_separator is not None)
    output = os.path.abspath(
        os.path.join(world.directory, "test", "k_fold1", "evaluation"))
    command = ("bigmler analyze --dataset " +
               world.dataset['resource'] +
               " --features --k-folds " + k_folds +
               " --output " + world.directory +
               " --exclude-features \"" + features + "\"" +
               " --args-separator " + args_separator +
               " --optimize " + metric)
    shell_execute(command, os.path.join(world.directory, "tmp"))
    world.output = output


def i_create_kfold_cross_validation_metric_category(
    step, k_folds=None, metric=None, category=None):
    """Step: I create BigML feature selection <k_folds>-fold cross-validations
    improving <metric> for category <category>
    """
    ok_(k_folds is not None and metric is not None and category is not None)
    output = os.path.abspath(
        os.path.join(world.directory, "test", "k_fold1", "evaluation"))
    command = ("bigmler analyze --dataset " +
               world.dataset['resource'] +
               " --features --k-folds " + k_folds +
               " --output " + world.directory +
               " --optimize " + metric +
               " --optimize-category " + category)
    shell_execute(command, os.path.join(world.directory, "tmp"))
    world.output = output


def i_create_kfold_cross_validation_metric(step, k_folds=None, metric=None):
    """Step: I create BigML feature selection <k_folds>-fold cross-validations
    improving <metric>
    """
    output = os.path.abspath(
        os.path.join(world.directory, "test", "k_fold1", "evaluation"))
    ok_(k_folds is not None and metric is not None)
    command = ("bigmler analyze --dataset " +
               world.dataset['resource'] +
               " --features --k-folds " + k_folds +
               " --output " + world.directory +
               " --optimize " + metric)
    shell_execute(command, os.path.join(world.directory, "tmp"))
    world.output = output


def i_generate_report(step):
    """Step: I generate a report from the output directory"""
    output = os.path.abspath(
        os.path.join(world.directory, "test", "k_fold1", "evaluation"))
    command = ("bigmler report --no-server --from-dir " +
               world.directory)
    shell_execute(command, os.path.join(world.directory, "tmp"))
    world.output = output


def is_symlink(step):
    """Step: a symlink file is generated in the reports directory"""
    resource_file = os.path.join(world.directory, REPORTS_DIR, "symlink")
    try:
        with open(resource_file, open_mode("r")) as resource_handler:
            ok_(resource_handler)
    except IOError as exc:
        ok_(False, msg=str(exc))


def retrieve_resource(scenario_path, resource_id):
    """Builds a resource from the contents of its local file

    """
    resource_file = os.path.join(scenario_path, resource_id.replace("/", "_"))
    with open(resource_file, open_mode("r")) as resource_handler:
        return json.loads(resource_handler.read())


def retrieve_resources(scenario_path, step):
    """Retrieves the resource ids and objects from the scenario path

    """
    try:
        with open(os.path.join(scenario_path, "source"), open_mode("r")) as \
                source_id_file:
            source_id = source_id_file.readline().strip()
            world.source = retrieve_resource(scenario_path, source_id)
    except Exception:
        pass
    try:
        with open(os.path.join(scenario_path, "dataset"), open_mode("r")) as \
                dataset_id_file:
            dataset_id = dataset_id_file.readline().strip()
            world.dataset = retrieve_resource(scenario_path, dataset_id)
    except Exception:
        pass
    try:
        with open(os.path.join(scenario_path, "source_test"), open_mode("r")) \
                as source_id_file:
            source_id = source_id_file.readline().strip()
            world.test_source = retrieve_resource(scenario_path, source_id)
    except Exception:
        pass
    try:
        with open(os.path.join(scenario_path, "dataset_test"),
                  open_mode("r")) as dataset_id_file:
            dataset_id = dataset_id_file.readline().strip()
            world.test_dataset = retrieve_resource(scenario_path, dataset_id)
    except Exception:
        pass
    try:
        with open(os.path.join(scenario_path, "models"), open_mode("r")) \
            as model_id_file:
            model_id = model_id_file.readline().strip()
            world.model = retrieve_resource(scenario_path, model_id)
    except Exception:
        pass
    try:
        with open(os.path.join(scenario_path, "deepnets"), open_mode("r")) \
            as model_id_file:
            model_id = model_id_file.readline().strip()
            world.deepnet = retrieve_resource(scenario_path, model_id)
    except Exception:
        pass
    try:
        with open(os.path.join(scenario_path, "logistic_regressions"),
                  open_mode("r")) as model_id_file:
            model_id = model_id_file.readline().strip()
            world.logistic_regression = retrieve_resource(scenario_path,
                                                          model_id)
    except Exception:
        pass
    try:
        with open(os.path.join(scenario_path, "linear_regressions"),
                  open_mode("r")) as model_id_file:
            model_id = model_id_file.readline().strip()
            world.linear_regression = retrieve_resource(scenario_path,
                                                        model_id)
    except Exception:
        pass


def i_have_previous_scenario_or_reproduce_it(step, scenario, kwargs):
    """Step: I have previously executed <scenario> or reproduce it with
    arguments <kwargs>
    """
    scenarios = {'scenario1': [(i_create_all_resources, True),
                               (i_check_create_source, False),
                               (i_check_create_dataset, False),
                               (i_check_create_model, False)],
                 'scenario1_img_dn': [(i_create_composite_source, True),
                                      (i_check_create_composite, False)],
                 'scenario1_r': [(i_create_all_resources, True),
                                 (i_check_create_source, False),
                                 (i_check_create_dataset, False),
                                 (i_check_create_model, False)],
                 'scenario5': [(i_create_resources_from_ensemble, True),
                               (i_check_create_ensemble, False)],
                 'scenario_e1': [(i_create_all_resources_to_evaluate, True),
                                 (i_check_create_source, False),
                                 (i_check_create_dataset, False),
                                 (i_check_create_model, False),
                                 (i_check_create_evaluation, False)],
                 'scenario_ml_1': [(i_create_all_ml_resources, True),
                                   (i_check_create_source, False),
                                   (i_check_create_dataset, False),
                                   (i_check_create_models, False)],
                 'scenario_ml_6': [(i_create_all_ml_resources, True),
                                   (i_check_create_source, False),
                                   (i_check_create_dataset, False),
                                   (i_check_create_models, False)],
                 'scenario_mle_1': [(i_create_all_ml_resources_and_ensembles,
                                     True),
                                    (i_check_create_source, False),
                                    (i_check_create_dataset, False),
                                    (i_check_create_models_in_ensembles,
                                     False)],
                 'scenario_ml_e1': [(i_create_all_ml_resources_for_evaluation,
                                     True),
                                    (i_check_create_source, False),
                                    (i_check_create_dataset, False),
                                    (i_check_create_models, False)],
                 'scenario_mc_1': [(i_create_all_mc_resources, True),
                                   (i_check_create_source, False),
                                   (i_check_create_dataset, False),
                                   (i_check_create_max_categories_datasets,
                                    False),
                                   (i_check_create_models, False)],
                 'scenario_r1': [(i_create_all_resources_batch, True),
                                 (i_check_create_source, False),
                                 (i_check_create_dataset, False),
                                 (i_check_create_model, False),
                                 (i_check_create_test_source, False),
                                 (i_check_create_test_dataset, False),
                                 (i_check_create_batch_prediction, False)],
                 'scenario_mlm_1': [(i_create_all_mlm_resources, True),
                                    (i_check_create_source, False),
                                    (i_check_create_dataset, False),
                                    (i_check_create_models, False)],
                'scenario_c_1': [(i_create_all_cluster_resources, True),
                                 (i_check_create_source, False),
                                 (i_check_create_dataset, False),
                                 (i_check_create_cluster, False)],
                'scenario_an_1': [(i_create_all_anomaly_resources, True),
                                  (i_check_create_source, False),
                                  (i_check_create_dataset, False),
                                  (i_check_create_anomaly, False)],
                'scenario_ass_1': [(i_create_association, True),
                                   (i_check_create_source, False),
                                   (i_check_create_dataset, False),
                                   (i_check_create_association, False)],
                'scenario1_lr': [(i_create_all_lr_resources, True),
                                 (i_check_create_source, False),
                                 (i_check_create_dataset, False),
                                 (i_check_create_lr_model, False)],
                'scenario1_dn': [(i_create_all_dn_resources,True),
                                 (i_check_create_source, False),
                                 (i_check_create_dataset, False),
                                 (i_check_create_dn_model, False)],
                'scenario1_pca': [(i_create_all_pca_resources, True),
                                  (i_check_create_source, False),
                                  (i_check_create_dataset, False),
                                  (i_check_create_pca_model, False)]}
    scenario_path = "%s/" % scenario
    if os.path.exists(scenario_path):
        retrieve_resources(scenario_path, step)
        assert True
    else:
        try:
            kwargs = json.loads(kwargs)
        except Exception as exc:
            print(str(exc))
        for function, with_args in scenarios[scenario]:
            if with_args:
                function(step, **kwargs)
            else:
                function(step)


def i_create_random_analysis(step, k_fold=None, metric=None):
    """Step: I create BigML random fields analysis with
    <k_fold>-cross-validation improving <metric>
    """
    output = os.path.abspath(
        os.path.join(world.directory, "test", "random1", "evaluation"))
    ok_(k_fold is not None and metric is not None)
    command = ("bigmler analyze --dataset " +
               world.dataset['resource'] +
               " --random-fields --number_of_models 2" +
               " --k-folds " + k_fold +
               " --output " + world.directory +
               " --optimize " + metric)
    shell_execute(command, os.path.join(world.directory, "tmp"))
    world.output = output


def i_check_create_kfold_random_forest(step, kfolds):
    """Step: I check that the <kfolds>-random forests have been created"""
    directory = os.path.dirname(os.path.dirname(world.output))
    directories = [os.path.join(directory, folder)
                   for folder in os.listdir(directory) if
                   os.path.isdir(os.path.join(directory, folder))]
    message = None
    for directory in directories:
        ensembles_file = os.path.join(directory, "ensembles")
        try:
            with open(ensembles_file) as handler:
                ensembles_list = list(
                    map(str.strip, handler.readlines()))
            world.ensembles.extend(ensembles_list)
            world.ensemble = ensembles_list[-1]
            eq_(int(kfolds), len(ensembles_list))
        except Exception as exc:
            message = str(exc)
            break
    ok_(message is None, msg=message)


def i_check_random_candidates(step, random_candidates, metric, metric_value):
    """Step: the best random candidates number is <random_candidates>,
    with <metric> of <metric_value>
    """
    ok_(random_candidates is not None and metric is not None and \
        metric_value is not None)
    sessions_file = os.path.join(world.directory, "bigmler_sessions")
    try:
        with open(sessions_file, open_mode("r")) as sessions_file:
            content = sessions_file.read()
        text = "The best random candidates number is: %s \n%s = %s" % (
            random_candidates, metric.capitalize(), metric_value)
        ok_(content.find(text) > -1)
    except Exception:
        ok_(False)



# -*- coding: utf-8 -*-
#
# Copyright 2020-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Resources management functions

"""


import sys
import os
import json

import bigml.api

from bigml.util import bigml_locale
from bigml.constants import TINY_RESOURCE

from bigmler.utils import (dated, get_url, log_message, check_resource,
                           check_resource_error, log_created_resources)

from bigmler.resourcesapi.common import set_basic_args, check_fields_struct, \
    update_attributes, update_json_args, set_config_attrs, hasattr_

from bigmler.resourcesapi.common import ALL_FIELDS_QS, LOCALE_DEFAULT
from bigmler.reports import report
from bigmler.processing.annotations import bigml_metadata


IMAGE_ATTRS = ["dimensions", "average_pixels", "level_histogram"]


def set_source_args(args, name=None, multi_label_data=None,
                    data_set_header=None, fields=None, update=False):
    """Returns a source arguments dict

    """

    if name is None:
        name = args.name
    source_args = set_basic_args(args, name)
    if args.project_id is not None:
        source_args.update({"project": args.project_id})
    # if header is set, use it
    if data_set_header is not None:
        source_args.update({"source_parser": {"header": data_set_header}})
    # If user has given an OS locale, try to add the locale used in bigml.com
    if args.user_locale is not None:
        source_locale = bigml_locale(args.user_locale)
        if source_locale is None:
            log_message("WARNING: %s locale equivalence not found."
                        " Using %s instead.\n" % (args.user_locale,
                                                  LOCALE_DEFAULT),
                        log_file=None, console=True)
            source_locale = LOCALE_DEFAULT
        source_args.update({'source_parser': {}})
        source_args["source_parser"].update({'locale': source_locale})
    # If user has set a training separator, use it.
    if args.training_separator is not None:
        training_separator = args.training_separator
        source_args["source_parser"].update({'separator': training_separator})
    # If uploading a multi-label file, add the user_metadata info needed to
    # manage the multi-label fields
    if (hasattr(args, 'multi_label') and args.multi_label
            and multi_label_data is not None):
        source_args.update({
            "user_metadata": {
                "multi_label_data": multi_label_data}})

    # image feature generation
    image_analysis = False
    extracted_features = []
    image_config = {}
    for attr in IMAGE_ATTRS:
        if hasattr_(args, attr) and getattr(args, attr):
            extracted_features.append(attr)
            image_analysis = True
    if hasattr_(args, "hog") and getattr(args, "hog"):
        extracted_features.append("histogram_of_gradients")
        image_analysis = True
    if hasattr_(args, "ws_level"):
        extracted_features.append(["wavelet_subbands", args.ws_level])
        image_analysis = True
    if hasattr_(args, "pretrained_cnn"):
        extracted_features.append(["pretrained_cnn", args.pretrained_cnn])
        image_analysis = True
    if extracted_features:
        args.image_analysis = True
    if hasattr_(args, "image_analysis"):
        image_config.update({"enabled": args.image_analysis})
        if extracted_features:
            image_config.update({"extracted_features": extracted_features})
        image_analysis = True
    if image_analysis:
        source_args.update({"image_analysis": image_config})

    if update:
        exclusive_attrs = ["add_sources_", "remove_sources_",
                           "delete_sources_", "closed", "sources_"]
        attr_aliases = {"add_sources_": "add_sources",
                        "remove_sources_": "remove_sources",
                        "delete_sources_": "delete_sources",
                        "sources_": "sources"}
        set_config_attrs(args, exclusive_attrs,
                         source_args, attr_aliases=attr_aliases,
                         exclusive=True)

        row_attrs = ["row_components", "row_indices", "row_values"]
        set_config_attrs(args, row_attrs, source_args)
    # to update fields attributes or types you must have a previous fields
    # structure (at update time)
    if fields:
        if args.field_attributes_:
            update_attributes(source_args,
                              {"fields": args.field_attributes_},
                              by_column=True, fields=fields)
        if args.types_:
            update_attributes(source_args,
                              {"fields": args.types_},
                              by_column=True, fields=fields)
        if args.import_fields:
            fields_struct = fields.new_fields_structure(args.import_fields)
            check_fields_struct(fields_struct, "source")
            update_attributes(source_args, fields_struct)
        if 'source' in args.json_args:
            update_json_args(source_args, args.json_args.get('source'), fields)
    return source_args


def create_source(data_set, source_args, args, api=None, path=None,
                  session_file=None, log=None, source_type=None):
    """Creates remote source

    """
    if api is None:
        api = bigml.api.BigML()
    suffix = "" if source_type is None else "%s " % source_type
    message = dated("Creating %ssource.\n" % suffix)
    log_message(message, log_file=session_file, console=args.verbosity)
    check_fields_struct(source_args, "source")
    if data_set is not None and not isinstance(data_set, str):
        # creating source from stream
        source = api.create_source(data_set, source_args)
    else:
        # annotated sources for images
        try:
            if os.path.exists(data_set) and not os.path.isdir(data_set):
                # --train metadata.json
                with open(data_set) as data_handler:
                    try:
                        source_info = json.load(data_handler)
                    except (UnicodeDecodeError, json.decoder.JSONDecodeError):
                        source_info = None
                if source_info is not None and isinstance(source_info, dict):
                    # could be a metadata.json file describing annotated images
                    source_attrs = source_info.keys()
                    if "annotations" in source_attrs and \
                            "images_file" in source_attrs:
                        source = api.create_annotated_source(data_set,
                                                             source_args)
                    else:
                        source = api.create_source(data_set, source_args)
                else:
                    # --train iris.csv
                    source = api.create_source(data_set, source_args)
            elif args.images_dir and args.annotations_file:
                # --train images_dir
                source = api.create_annotated_source(bigml_metadata(args),
                                                     source_args)
            else:
                source = api.create_source(data_set, source_args)
        except (IOError, TypeError):
            # empty composite source, for instance
            source = api.create_source(args.sources_, source_args)
    if path is not None:
        suffix = "_" + source_type if source_type else ""
        log_created_resources(
            "source%s" % suffix, path,
            source['resource'], mode='ab',
            comment=("%s\n" % source.get('object', {}).get('name')))
    source_id = check_resource_error(source, "Failed to create source: ")
    try:
        source = check_resource(source, api.get_source,
                                query_string=ALL_FIELDS_QS,
                                raise_on_error=True)
    except Exception as exception:
        sys.exit("Failed to get a finished source: %s" % str(exception))
    message = dated("Source created: %s\n" % get_url(source))
    log_message(message, log_file=session_file, console=args.verbosity)
    log_message("%s\n" % source_id, log_file=log)
    if hasattr(args, "reports") and args.reports:
        report(args.reports, path, source)
    if hasattr(source_args, "sources_"):
        # if we used the list of sources to create a composite, we don't
        # want the argument to be used in the update step
        args.sources_ = None
    return source


def clone_source(source, args, api=None, path=None,
                 session_file=None, log=None):
    """Clones the source to open it for editing

    """
    if api is None:
        api = bigml.api.BigML()
    message = dated("Cloning source to open it for editing\n")
    log_message(message, log_file=session_file, console=args.verbosity)
    source = api.clone_source(source)
    source_id = check_resource_error(source, "Failed to clone source: ")
    try:
        source = check_resource(source, api.get_source,
                                query_string=ALL_FIELDS_QS,
                                raise_on_error=True)
    except Exception as exception:
        sys.exit("Failed to get a finished source: %s" % str(exception))
    log_created_resources("source", path, source['resource'], mode='ab',
        comment=f"{source['object']['name']}\n")
    message = dated("Source created: %s\n" % get_url(source))
    log_message(message, log_file=session_file, console=args.verbosity)
    log_message("%s\n" % source_id, log_file=log)
    return source


def data_to_source(args):
    """Extracts the flags info to create a source object

    """
    data_set = None
    data_set_header = None
    if (args.training_set and (not args.source or args.annotations_file) and not
            (hasattr(args, "dataset") and args.dataset) and
            not args.has_models_):
        data_set = args.training_set
        data_set_header = args.train_header
    elif (hasattr(args, 'evaluate') and args.evaluate and args.test_set
          and not args.source):
        data_set = args.test_set
        data_set_header = args.test_header
    if data_set is not None:
        try:
            with open(data_set, 'r') as data_handler:
                content = json.load(data_handler)
                if isinstance(content, dict) and \
                        "externalconnector_id" in content \
                        and "query" in content:
                    data_set = content
        except (IOError, ValueError, TypeError):
            pass
    return data_set, data_set_header


def get_source(source, api=None, verbosity=True,
               session_file=None):
    """Retrieves the source in its actual state and its field info

    """
    if api is None:
        api = bigml.api.BigML()
    if (isinstance(source, str) or
            bigml.api.get_status(source)['code'] != bigml.api.FINISHED):
        message = dated("Retrieving source. %s\n" %
                        get_url(source))
        log_message(message, log_file=session_file,
                    console=verbosity)
        try:
            source = check_resource(source, api.get_source,
                                    query_string=ALL_FIELDS_QS,
                                    raise_on_error=True)
        except Exception as exception:
            sys.exit("Failed to get a finished source: %s" % str(exception))

    return source


def update_source(source, source_args, args,
                  api=None, verbosity=True, session_file=None, log=None):
    """Updates source properties

    """
    if api is None:
        api = bigml.api.BigML()
    message = dated("Updating source. %s\n" %
                    get_url(source))
    log_message(message, log_file=session_file,
                console=args.verbosity)
    source_min = api.get_source(source, query_string=TINY_RESOURCE)
    if source_min["object"].get("closed", False):
        message = dated("Source was closed. Cloning %s\n" %
                        get_url(source))
        log_message(message, log_file=session_file,
                    console=verbosity)
        source = api.clone_source(source)
        message = "Source created: %s\n" % get_url(source)
        log_message(message, log_file=session_file, console=args.verbosity)
        log_message("%s\n" % source["object"]["resource"], log_file=log)

    source = api.update_source(source, source_args)
    check_resource_error(source, "Failed to update source: ")
    source = check_resource(source, api.get_source)
    return source



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2014-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from bigmler.tests.common_steps import shell_execute, base_shell_execute
from bigmler.tests.world import world, res_filename, ok_, eq_


def i_create_all_mlm_resources(
    step, tag=None, label_separator=None, number_of_labels=None, data=None,
    training_separator=None, ml_fields=None, model_fields=None, objective=None,
    test=None, output=None):
    """Step: I create BigML multi-label resources tagged as <tag> with
    <label_separator> label separator and <number_of_labels> labels uploading
    train <data> file with <training_separator> field separator and
    <ml_fields> as multi-label fields using model_fields <model_fields>
    and objective <objective> to test <test> and log predictions in <output>
    """
    ok_(tag is not None and label_separator is not None and
        training_separator is not None and number_of_labels is not None
        and data is not None and test is not None and output is not None
        and model_fields is not None and objective is not None and
        ml_fields is not None)
    world.number_of_models = int(number_of_labels)
    test = res_filename(test)
    command = ("bigmler --multi-label --train " + res_filename(data) +
               " --multi-label-fields " + ml_fields +
               " --label-separator \"" + label_separator +
               "\" --training-separator \"" + training_separator +
               "\" --model-fields \" " + model_fields +
               "\" --test " + test + " --store --output " + output +
               " --objective " + objective +
               " --tag " + tag + " --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_create_all_ml_resources(
    step, tag=None, label_separator=None, number_of_labels=None, data=None,
    training_separator=None, test=None, output=None):
    """Step: I create BigML multi-label resources tagged as <tag> with
    <label_separator> label separator and <number_of_labels> labels uploading
    train <data> file with <training_separator> field separator to test <test>
    and log predictions in <output>
    """
    ok_(tag is not None and label_separator is not None and
        training_separator is not None and number_of_labels is not None and
        data is not None and test is not None and output is not None)
    world.number_of_models = int(number_of_labels)
    test = res_filename(test)
    command = ("bigmler --multi-label --train " + res_filename(data) +
               " --label-separator \"" + label_separator +
               "\" --training-separator \"" + training_separator +
               "\" --test " + test + " --store --output " + output +
               " --tag " + tag + " --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_predict_ml_from_model_tag(step, tag=None, test=None, output=None):
    """Step: I create BigML multi-label resources using models tagged as <tag>
    to test <test> and log predictions in <output>
    """
    ok_(tag is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler --multi-label --model-tag " + tag + " --test " +
               test + " --store --output " + output +
               " --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_predict_ml_from_model_tag_with_labels_with_objective(
    step, labels=None, objective=None, tag=None, test=None, output=None):
    """Step: I create BigML multi-label resources with labels <labels>
    and objective <objective> using models tagged as <tag> to test <test>
    and log predictions in <output>
    """
    ok_(tag is not None and labels is not None and test is not None and
        output is not None and objective is not None)
    test = res_filename(test)
    command = ("bigmler --multi-label --model-tag " + tag + " --labels " +
               labels + " --test " + test + " --store --output " + output +
               " --objective " + objective + " --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_predict_ml_from_model_tag_with_labels(
    step, labels=None, tag=None, test=None, output=None):
    """Step: I create BigML multi-label resources with labels <labels>
    using models tagged as <tag> to test <test> and log predictions in
    <output>
    """
    ok_(tag is not None and labels is not None and test is not None and
        output is not None)
    test = res_filename(test)
    command = ("bigmler --multi-label --model-tag " + tag + " --labels " +
               labels + " --test " + test + " --store --output " + output +
               " --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_check_local_file(step, path=None):
    """Step: Then I check the extended file <path> has been created"""
    ok_(path is not None)
    message = None
    try:
        with open(path) as handler:
            world.headers = handler.readline().strip()
            world.first_row = handler.readline().strip()
    except IOError as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_headers_file(step, headers=None):
    """Step: the headers of the local extended file are <headers>"""
    ok_(headers is not None)
    eq_(headers, world.headers, msg="The expected headers are:\n%s\nand the"
        " ones found are:\n%s" % (headers, world.headers))


def i_check_first_row_file(step, first_row=None):
    """Step: the first row of the local extended file is <first_row>"""
    ok_(first_row is not None)
    eq_(first_row, world.first_row, msg=("The expected first row is:\n"
        f"{first_row}\nand the one found is:\n{world.first_row}"))


def i_create_ml_source(
    step, label_separator=None, number_of_labels=None, data=None,
    training_separator=None, multi_label_fields=None, objective=None,
    output_dir=None):
    """Step: I create BigML a multi-label source with <label_separator> label
    separator and <number_of_labels> labels from train <data> file with
    <training_separator> field separator and <multi_label_fields> as
    multi-label fields and objective <objective> and output in <output_dir>
    """
    ok_(label_separator is not None and training_separator is not None and
        number_of_labels is not None and data is not None and
        multi_label_fields is not None and output_dir is not None and
        objective is not None)
    command = ("bigmler --multi-label --train " + res_filename(data) +
               " --label-separator \"" + label_separator +
               "\" --training-separator \"" + training_separator +
               "\" --multi-label-fields " + multi_label_fields +
               " --objective " + objective + " --store --output-dir " +
               output_dir +
               "  --no-dataset --no-model --max-batch-models 1")
    base_shell_execute(command, output_dir)


def i_create_all_ml_resources_and_ensembles(
    step, tag=None, label_separator=None, number_of_labels=None, data=None,
    training_separator=None, number_of_models=None, test=None, output=None):
    """Step: I create BigML multi-label resources tagged as <tag> with
    <label_separator> label separator and <number_of_labels> labels uploading
    train <data> file with <training_separator> field separator and
    <number_of_models> models ensembles to test <test> and log predictions
    in <output>
    """
    ok_(tag is not None and label_separator is not None and
        training_separator is not None and number_of_labels is not None and
        data is not None and test is not None and output is not None and
        number_of_models is not None)
    world.number_of_models = int(number_of_labels) * int(number_of_models)
    test = res_filename(test)
    command = ("bigmler --multi-label --train " + res_filename(data) +
               " --label-separator \"" + label_separator +
               "\" --training-separator \"" + training_separator +
               "\" --test " + test + " --number-of-models " +
               str(number_of_models) + " --store --output " + output +
               " --tag " + tag + " --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_create_resources_and_ensembles_from_source(
    step, multi_label=None, number_of_models=None, test=None, output=None):
    """Step: I create BigML <multi-label> resources using source and
    <number_of_models> models ensembles to test <test> and log predictions
    in <output>
    """
    ok_(test is not None and output is not None and
        number_of_models is not None)
    multi_label = "" if multi_label is None else " --multi-label "
    test = res_filename(test)
    command = ("bigmler "+ multi_label +"--source " +
               world.source['resource'] + " --number-of-models " +
               str(number_of_models) + " --test " + test +
                " --store --output " + output)
    shell_execute(command, output, test=test)


def i_create_resources_and_ensembles_from_dataset(
    step, multi_label=None, number_of_models=None, test=None, output=None):
    """Step: I create BigML <multi-label> resources using dataset and
    <number_of_models> models ensembles to test <test> and log predictions
    in <output>
    """
    ok_(test is not None and output is not None)
    multi_label = "" if multi_label is None else " --multi-label "
    test = res_filename(test)
    command = ("bigmler "+ multi_label +"--dataset " +
               world.dataset['resource'] + " --number-of-models " +
               str(number_of_models) + " --test " + test +
               " --store --output " + output)
    shell_execute(command, output, test=test)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2014-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import os
import json
import sys

from bigml.api import check_resource

from bigmler.tests.common_steps import base_shell_execute
from bigmler.tests.world import world, res_filename, ok_, eq_


def i_create_dataset(step, data=None, output_dir=None):
    """Step: I create a BigML dataset from <data> and store logs in
    <output_dir>
    """
    ok_(data is not None and output_dir is not None)
    command = ("bigmler --train " + res_filename(data) +
               " --no-model --store --output-dir " + output_dir)
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)


def i_create_dataset_from_source(step, output_dir=None):
    """Step: I create a BigML dataset from previous source and store logs in
    <output_dir>
    """
    ok_(output_dir is not None)
    command = (("bigmler --source %s" % world.source['resource']) +
               " --no-model --store --output-dir " + output_dir)
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)


def i_create_filtered_dataset_from_dataset(
    step, filter_exp=None, output_dir=None):
    """Step: I create a BigML filtered dataset with filter <filter_exp>
    from previous dataset and store logs in <output_dir>
    """
    ok_(filter_exp is not None and output_dir is not None)
    if sys.platform != 'win32':
        filter_exp = '"%s"' % filter_exp.replace('"', '\\"')
    command = ('echo ' +
               filter_exp + ' > ' +
               output_dir + "/filter.lisp")
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)
    command = (("bigmler --dataset %s" % world.dataset['resource']) +
               " --no-model --store --output-dir " + output_dir +
               " --lisp-filter " + output_dir + "/filter.lisp")
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)


def i_create_dataset_new_fields(step, json_file=None, model_fields=None):
    """Step: I create a new BigML dataset using the specs in JSON file
    <json_file> and a model with <model_fields>
    """
    ok_(json_file is not None and model_fields is not None)
    json_file = res_filename(json_file)
    command = ("bigmler --dataset " + world.dataset['resource'] +
               " --model-fields \"" + model_fields + "\" --store" +
               " --output-dir " + world.directory +
               " --new-fields " + json_file)
    message = base_shell_execute(command, world.directory)
    ok_(message is None, msg=message)


def i_update_dataset_new_properties(step, json_file=None):
    """Step: I update the dataset using the specs in JSON file <json_file>"""
    ok_(json_file is not None)
    json_file = res_filename(json_file)
    command = ("bigmler --dataset " + world.dataset['resource'] +
               " --no-model --store --output-dir " + world.directory +
               " --dataset-attributes " + json_file)
    message = base_shell_execute(command, world.directory)
    ok_(message is None, msg=message)


def i_check_dataset_has_field(step, field):
    """Step: I check that the new dataset has field <field>"""
    fields = world.dataset['object']['fields']
    for field_id in fields:
        if fields[field_id]['name'] == field:
            ok_(True)
            return
    ok_(False)


def i_check_dataset_has_property(
    step, attribute=None, field_id=None, value=None, attr_type=None):
    """Step: I check that property <attribute> for field id <field_id> is
    <value> of type <type>
    """
    if attribute is None or field_id is None or value is None:
        assert False
    fields = world.dataset['object']['fields']
    if attr_type == "boolean":
        value = bool(value)
    if attr_type == "integer":
        value = int(value)
    if fields[field_id][attribute] == value:
        ok_(True)
        return
    ok_(False)


def i_create_multi_dataset(step, output_dir):
    """Step: I create a multi-dataset from the datasets file and store logs
    in <output_dir>
    """
    ok_(output_dir is not None)
    datasets_file = os.path.join(world.directory, "dataset")
    command = ("bigmler --datasets " + datasets_file +
               " --multi-dataset --no-model --store --output-dir " +
               output_dir)
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)


def i_check_multi_dataset_origin(step, output_dir=None):
    """Step: I check that the multi-dataset\'s origin are the datasets in
    <output_dir>
    """
    ok_(output_dir is not None)
    datasets_file = os.path.join(output_dir, "dataset")
    message = None
    try:
        origin_datasets = world.dataset['object']['origin_datasets']
        count = 0
        with open(datasets_file) as datasets_file_handler:
            for dataset_id in datasets_file_handler:
                dataset_id = dataset_id.strip()
                ok_(dataset_id in origin_datasets,
                    msg=f"{dataset_id} not in {origin_datasets}")
                count += 1
        eq_(count, len(origin_datasets))
    except KeyError as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_filter_field_from_dataset(step, field=None, output_dir=None):
    """Step: I filter out field <field> from dataset and log to <output_dir>"""
    ok_(field is not None and output_dir is not None)
    empty_json = res_filename('data/empty.json')
    command = ("bigmler --dataset " + world.dataset['resource'] +
               " --no-model --store --output-dir " + output_dir +
               " --dataset-fields=\"-" + field + "\""+
               " --new-fields " + empty_json)
    message = base_shell_execute(command, world.directory)
    ok_(message is None, msg=message)


def i_check_create_multi_dataset(step):
    """Step: I check that the multi-dataset has been created"""
    dataset_file = os.path.join(world.directory, "dataset_multi")
    message = None
    try:
        with open(dataset_file) as dataset_file_handler:
            dataset_id = dataset_file_handler.readline().strip()
        dataset = check_resource(dataset_id,
                                 world.api.get_dataset)
        world.datasets.append(dataset['resource'])
        world.dataset = dataset
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_files_equal(step, local_file, data):
    """Step: file <local_file> is like file <data>"""
    data = res_filename(data)
    with open(os.path.join(world.directory, local_file)) as handler:
        contents_local = handler.read()
    with open(data) as handler:
        contents = handler.read()
    eq_(contents_local, contents)


def i_export_the_dataset(step, filename):
    """Step: I export the dataset to the CSV file <filename>"""
    ok_(filename is not None)
    command = ("bigmler --dataset " + world.dataset['resource'] +
               " --to-csv " + filename +
               " --output-dir " + world.directory + " --no-model")
    message = base_shell_execute(command, world.directory)
    ok_(message is None, msg=message)


def i_create_dataset_with_summary(
    step, data=None, summary_file=None, output_dir=None):
    """Step: I create a BigML dataset from <data> and a summary file
    <summary_file> for its fields and store logs in <output_dir>
    """
    ok_(data is not None and output_dir is not None and
        summary_file is not None)
    command = ("bigmler --train " + res_filename(data) +
               " --no-model --store --output-dir " + output_dir +
               " --export-fields " + summary_file)
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)


def i_create_juxtaposed(step, output_dir=None):
    """Step: I create a new dataset juxtaposing both datasets"""
    ok_(output_dir is not None)
    command = ("bigmler dataset --datasets " +
               os.path.join(output_dir, "dataset") +
               " --juxtapose --store --output-dir " + output_dir)
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)


def i_create_join(step, output_dir=None, sql=None):
    """Step: I create a new dataset joining both datasets"""
    ok_(output_dir is not None and sql is not None)
    command = ("bigmler dataset --datasets " +
               os.path.join(output_dir, "dataset") +
               " --sql-query \"" + sql + "\" --store --output-dir " +
               output_dir)
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)


def i_create_cluster_with_params_from_dataset(
    step, cluster_params=None, output_dir=None):
    """Step: I create a BigML cluster with params <cluster_params> from dataset
    in <output_dir>
    """
    ok_(output_dir is not None and cluster_params is not None)
    command = ("bigmler cluster --dataset " + world.dataset['resource'] +
               " --store --output-dir " + output_dir +
               " " + cluster_params)
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)


def i_create_anomaly_with_params_from_dataset(
    step, params=None, output_dir=None):
    """Step: I create a BigML anomaly with params <params> from dataset in
    <output_dir>
    """
    ok_(output_dir is not None and params is not None)
    command = ("bigmler anomaly --dataset " + world.dataset['resource'] +
               " --store --output-dir " + output_dir +
               " " + params)
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)


def i_create_logistic_with_params_from_dataset(
    step, params=None, output_dir=None):
    """Step: I create a BigML logistic with params <params> from dataset in
    <output_dir>
    """
    ok_(output_dir is not None and params is not None)
    command = ("bigmler logistic-regression --dataset " +
               world.dataset['resource'] +
               " --store --output-dir " + output_dir +
               " " + params)
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)


def i_create_association_with_params_from_dataset(
    step, params=None, output_dir=None):
    """Step: I create a BigML association with params <params> from dataset in
    <output_dir>
    """
    ok_(output_dir is not None and params is not None)
    command = ("bigmler association --dataset " +
               world.dataset['resource'] +
               " --store --output-dir " + output_dir +
               " " + params)
    message = base_shell_execute(command, output_dir)
    ok_(message is None, msg=message)


def i_import_fields(step, summary=None):
    """Step: I import fields attributes in file <summary> to dataset"""
    ok_(summary is not None)
    command = ("bigmler --dataset " + world.dataset['resource'] +
               " --import-fields " + res_filename(summary) +
               " --output-dir " + world.directory + " --no-model")
    message = base_shell_execute(command, world.directory)
    ok_(message is None, msg=message)


def field_attribute_value(step, field=None, attribute=None,
                          attribute_value=None):
    """Step: the field <field> has <attribute> equal to <attribute_value>"""
    dataset = check_resource(world.dataset['resource'],
                             world.api.get_dataset)
    fields = dataset['object']['fields']
    eq_(fields[field][attribute], attribute_value)


def i_check_cluster_params(step, params_json=None):
    """Step: the cluster params are <params_json>"""
    params_dict = json.loads(params_json)
    for key, value in list(params_dict.items()):
        eq_(value, world.cluster['object'].get(key))


def i_check_anomaly_params(step, params_json=None):
    """Step: the anomaly params are <params_json>"""
    params_dict = json.loads(params_json)
    for key, value in list(params_dict.items()):
        eq_(value, world.anomaly['object'].get(key))


def i_check_logistic_params(step, params_json=None):
    """Step: the logistic params are <params_json>"""
    params_dict = json.loads(params_json)
    for key, value in list(params_dict.items()):
        eq_(value, world.logistic_regression['object'].get(key))


def i_check_association_params(step, params_json=None):
    """Step: the association params are <params_json>"""
    params_dict = json.loads(params_json)
    for key, value in list(params_dict.items()):
        eq_(value, world.association['object'].get(key))


def i_check_juxtaposed(step, datasets=None):
    """Step: I check that datasets have been juxtaposed"""
    rows = world.dataset["object"]["rows"]
    eq_(rows, datasets[0]["object"]["rows"])
    number_of_fields = 0
    for dataset in datasets:
        number_of_fields += dataset["object"]["field_types"]["total"]
    eq_(number_of_fields, world.dataset["object"]["field_types"]["total"])


def i_check_joined(step, rows=None):
    """Step: I check that datasets have been joined"""
    eq_(rows, world.dataset["object"]["rows"])



# -*- coding: utf-8 -*-
def predict_imagen(data={}):
    """ Predictor for Imagen

        Created using BigMLer
    """

    import re

    tm_tokens = 'tokens_only'
    tm_full_term = 'full_terms_only'
    tm_all = 'all'

    def term_matches(text, field_name, term):
        """ Counts the number of occurences of term and its variants in text

        """
        forms_list = term_forms[field_name].get(term, [term])
        options = term_analysis[field_name]
        token_mode = options.get('token_mode', tm_tokens)
        case_sensitive = options.get('case_sensitive', False)
        first_term = forms_list[0]
        if token_mode == tm_full_term:
            return full_term_match(text, first_term, case_sensitive)
        else:
            # In token_mode='all' we will match full terms using equals and
            # tokens using contains
            if token_mode == tm_all and len(forms_list) == 1:
                pattern = re.compile(r'^.+\b.+$', re.U)
                if re.match(pattern, first_term):
                    return full_term_match(text, first_term, case_sensitive)
            return term_matches_tokens(text, forms_list, case_sensitive)


    def full_term_match(text, full_term, case_sensitive):
        """Counts the match for full terms according to the case_sensitive
              option

        """
        if not case_sensitive:
            text = text.lower()
            full_term = full_term.lower()
        return 1 if text == full_term else 0

    def get_tokens_flags(case_sensitive):
        """Returns flags for regular expression matching depending on text
              analysis options

        """
        flags = re.U
        if not case_sensitive:
            flags = (re.I | flags)
        return flags


    def term_matches_tokens(text, forms_list, case_sensitive):
        """ Counts the number of occurrences of the words in forms_list in
               the text

        """
        flags = get_tokens_flags(case_sensitive)
        expression = r'(\b|_)%s(\b|_)' % '(\\b|_)|(\\b|_)'.join(forms_list)
        pattern = re.compile(expression, flags=flags)
        matches = re.findall(pattern, text)
        return len(matches)

    term_analysis = {
        "titulo": {
            "case_sensitive": False,
            "token_mode": 'all',
        },
    }
    term_forms = {
        "titulo": {
            "fantásticos": ['fantásticos', 'fantásticas'],
            "gigante": ['gigante', 'gigantes'],
        },
    }

    if (data.get('codbarras') is None):
        return {"prediction": 1.82, "error": 5.53698}
    if (data['codbarras'] > 9789872414340):
        if (data.get('ano_lanzamiento') is None):
            return {"prediction": 9, "error": 7.02326}
        if (data['ano_lanzamiento'] > 2008):
            if (data.get('paginas') is None):
                return {"prediction": 10.5, "error": 5.88884}
            if (data['paginas'] > 90):
                if (data.get('titulo') is None):
                    return {"prediction": 9, "error": 5.08228}
                if (term_matches(data['titulo'], "titulo", u"fantásticos") > 0):
                    return {"prediction":8, "error":5.08228}
                if (term_matches(data['titulo'], "titulo", u"fantásticos") <= 0):
                    if (data.get('grados') is None):
                        return {"prediction": 9.5, "error": 5.26764}
                    if (data['grados'] == "Elsa Pizzi"):
                        return {"prediction":9, "error":5.26764}
                    if (data['grados'] != "Elsa Pizzi"):
                        return {"prediction":10, "error":5.26764}
            if (data['paginas'] <= 90):
                if (data.get('titulo') is None):
                    return {"prediction": 12, "error": 5.08228}
                if (term_matches(data['titulo'], "titulo", u"gigante") > 0):
                    return {"prediction":11, "error":5.08228}
                if (term_matches(data['titulo'], "titulo", u"gigante") <= 0):
                    if (data.get('grados') is None):
                        return {"prediction": 12.5, "error": 5.26764}
                    if (data['grados'] == "Patricia Roggio"):
                        return {"prediction":13, "error":5.26764}
                    if (data['grados'] != "Patricia Roggio"):
                        return {"prediction":12, "error":5.26764}
        if (data['ano_lanzamiento'] <= 2008):
            if (data.get('grados') is None):
                return {"prediction": 6, "error": 5.08228}
            if (data['grados'] == "4°, 5°"):
                return {"prediction":7, "error":5.08228}
            if (data['grados'] != "4°, 5°"):
                if (data['grados'] == "5°, 6°"):
                    return {"prediction":5, "error":5.26764}
                if (data['grados'] != "5°, 6°"):
                    return {"prediction":6, "error":5.26764}
    if (data['codbarras'] <= 9789872414340):
        if (data['codbarras'] > 9789872414309):
            if (data.get('paginas') is None):
                return {"prediction": 3, "error": 5.08228}
            if (data['paginas'] > 100):
                if (data.get('grados') is None):
                    return {"prediction": 2.5, "error": 5.26764}
                if (data['grados'] == "4°, 5°"):
                    return {"prediction":2, "error":5.26764}
                if (data['grados'] != "4°, 5°"):
                    return {"prediction":3, "error":5.26764}
            if (data['paginas'] <= 100):
                return {"prediction":4, "error":5.08228}
        if (data['codbarras'] <= 9789872414309):
            if (data['codbarras'] > 9789871989852):
                return {"prediction":1, "error":0.26071}
            if (data['codbarras'] <= 9789871989852):
                return {"prediction":0, "error":0.04286}


def predict(data={}):
    prediction = predict_imagen(data=data)
    return prediction


# -*- coding: utf-8 -*-
#
# Copyright 2020-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Resources management functions

"""


import sys

import bigml.api

from bigmler.utils import (dated, get_url, log_message, check_resource,
                           check_resource_error, log_created_resources)

from bigmler.resourcesapi.common import set_basic_args, \
    update_attributes


def set_script_args(args, name=None):
    """Returns a script arguments dict

    """

    if name is None:
        name = args.name
    script_args = set_basic_args(args, name)
    if args.project_id is not None:
        script_args.update({"project": args.project_id})
    if args.imports is not None:
        script_args.update({"imports": args.imports_})
    if args.parameters_ is not None:
        script_args.update({"inputs": args.parameters_})
    if args.declare_outputs_:
        script_args.update({"outputs": args.declare_outputs_})
    update_attributes(script_args, args.json_args.get('script'))
    return script_args


def create_script(source_code, script_args, args, api=None, path=None,
                  session_file=None, log=None):
    """Creates remote script

    """

    if api is None:
        api = bigml.api.BigML()
    message = dated("Creating script \"%s\".\n" % script_args["name"])
    log_message(message, log_file=session_file, console=args.verbosity)
    script = api.create_script(source_code, script_args)
    log_created_resources("scripts", path,
                          bigml.api.get_script_id(script), mode='a')
    script_id = check_resource_error(script, "Failed to create script: ")
    try:
        script = check_resource(script, api.get_script, raise_on_error=True)
    except Exception as exception:
        sys.exit("Failed to get a compiled script: %s" % str(exception))
    message = dated("Script created: %s\n" % get_url(script))
    log_message(message, log_file=session_file, console=args.verbosity)
    log_message("%s\n" % script_id, log_file=log)
    return script


def get_script(script, api=None, verbosity=True,
               session_file=None):
    """Retrieves the script in its actual state

    """
    if api is None:
        api = bigml.api.BigML()
    if (isinstance(script, str) or
            bigml.api.get_status(script)['code'] != bigml.api.FINISHED):
        message = dated("Retrieving script. %s\n" %
                        get_url(script))
        log_message(message, log_file=session_file,
                    console=verbosity)
        try:
            script = check_resource(script, api.get_script,
                                    raise_on_error=True)
        except Exception as exception:
            sys.exit("Failed to get a compiled script: %s" % str(exception))
    return script



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2014-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from bigmler.tests.common_steps import shell_execute, base_shell_execute
from bigmler.tests.world import res_filename, ok_

CAT = "type " if sys.platform == "win32" else "cat "


def i_create_all_resources_to_test_from_stdin(
    step, data=None, test=None, name=None, output=None):
    """Step: I create BigML resources uploading train <data> file to test
    <test> read from stdin with name <name> and log predictions in <output>
    """
    ok_(data is not None and test is not None and output is not None and
        name is not None)
    test = res_filename(test)

    command = (CAT + test + "|bigmler --train " + res_filename(data) +
               " --test --store --output " + output + " --name \"" +
               name + "\" --max-batch-models 1")
    shell_execute(command, output, test=test)


def i_create_source_from_stdin(step, data=None, output_dir=None):
    """Step: I create a BigML source from stdin using train <data>
    file and logging in <output_dir>
    """
    ok_(data is not None and output_dir is not None)
    command = (CAT + res_filename(data) + "|bigmler --train " +
               "--store --no-dataset --no-model --output-dir " +
               output_dir + " --max-batch-models 1")
    base_shell_execute(command, output_dir)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2016-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import os

from bigml.api import check_resource

from bigmler.tests.common_steps import shell_execute
from bigmler.tests.world import world, res_filename, ok_


def i_create_all_lr_resources_with_no_headers(
    step, data=None, test=None, output=None):
    """Step: I create BigML resources uploading train <data> file with no
    headers to test <test> with no headers and log predictions in <output>
    """
    ok_(data is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler logistic-regression --train " + res_filename(data) +
               " --test " + test +
               " --store --no-balance-fields --no-bias --output " + output +
               " --no-train-header --no-test-header")
    shell_execute(command, output, test=test)


def i_create_all_lr_resources(step, data=None, test=None, output=None):
    """Step: I create BigML resources uploading train <data> file to test
    <test> and log predictions in <output>
    """
    ok_(data is not None and test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler logistic-regression --train " + res_filename(data) +
               " --test " + test +
               " --store --no-balance-fields --no-bias --output " + output)
    shell_execute(command, output, test=test)


def i_create_lr_resources_from_source(
    step, multi_label=None, test=None, output=None):
    """Step: I create BigML <multi-label> resources using source to test
    <test> and log predictions in <output>
    """
    ok_(test is not None and output is not None)
    test = res_filename(test)
    multi_label = "" if multi_label is None else " --multi-label "
    command = ("bigmler logistic-regression "+ multi_label +"--source " +
               world.source['resource'] + " --test " + test +
               " --store --no-bias --no-balance-fields --output " + output)
    shell_execute(command, output, test=test)


def i_create_lr_resources_from_dataset(
    step, multi_label=None, test=None, output=None):
    """Step: I create BigML <multi-label> resources using dataset to test
    <test> and log predictions in <output>
    """
    ok_(test is not None and output is not None)
    multi_label = "" if multi_label is None else " --multi-label "
    test = res_filename(test)
    command = ("bigmler logistic-regression "+ multi_label +"--dataset " +
               world.dataset['resource'] + " --test " + test +
               " --store --no-balance-fields --no-bias --output " + output)
    shell_execute(command, output, test=test)


def i_create_lr_resources_from_model(step, test=None, output=None):
    """Step: I create BigML resources using model to test <test> and log
    predictions in <output>
    """
    ok_(test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler logistic-regression --logistic-regression " +
               world.logistic_regression['resource'] + " --test " +
               test + " --store --no-balance-fields --no-bias --output " +
               output)
    shell_execute(command, output, test=test)


def i_create_lr_resources_from_model_remote(step, test=None, output=None):
    """Step: I create BigML resources using model to test <test> as batch
    prediction and log predictions in <output>
    """
    ok_(test is not None and output is not None)
    test = res_filename(test)
    command = ("bigmler logistic-regression --logistic-regression " +
               world.logistic_regression['resource'] + " --test " + test +
               " --store --no-balance-fields " +
               " --no-bias --remote --output " +
               output)
    shell_execute(command, output, test=test, options="--no-header")


def i_create_lr_resources_from_model_with_op(step, test=None, output=None,
                                             operating_point=None):
    """Step: I create BigML resources using model to test <test> as batch
    prediction and log predictions in <operating_point>
    """
    ok_(test is not None and output is not None and \
        operating_point is not None)
    test = res_filename(test)
    operating_point = res_filename(operating_point)
    command = ("bigmler logistic-regression --logistic-regression " +
               world.logistic_regression['resource'] + " --test " + test +
               " --operating-point " + operating_point +
               " --store --no-balance-fields --no-bias --output " +
               output)
    shell_execute(command, output, test=test)


def i_check_create_lr_model(step):
    """Step: I check that the logistic regression model has been created"""
    lr_file = os.path.join(world.directory, "logistic_regressions")
    message = None
    try:
        with open(lr_file) as handler:
            logr = check_resource(handler.readline().strip(),
                                  world.api.get_logistic_regression)
            world.logistic_regressions.append(logr['resource'])
            world.logistic_regression = logr
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_create_all_lr_resources_to_evaluate(step, data=None, output=None):
    """Step: I create BigML logistic regression resources uploading train
    <train> file to evaluate and log evaluation in <output>
    """
    ok_(data is not None and output is not None)
    command = ("bigmler logistic-regression --train " + res_filename(data) +
               " --evaluate" +
               " --store --no-balance-fields --no-bias --output " + output)
    shell_execute(command, output)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2016-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.



import os
import json

from bigml.api import check_resource

from bigmler.tests.common_steps import shell_execute
from bigmler.tests.world import world, res_filename, ok_, eq_


def i_create_all_execution_resources(step, code=None, output_dir=None):
    """Step: I create BigML execution resources from source code <code> and log
    results in <output_dir>
    """
    ok_(code is not None and output_dir is not None)
    command = ("bigmler execute --code \"" + code +
               "\" --store --output-dir " + output_dir)
    shell_execute(command, "%s/xx.txt" % output_dir)


def i_create_all_library_resources(step, code_file=None, output_dir=None):
    """Step: I create BigML library from code in file <code_file> and log
    results in <output_dir>
    """
    ok_(code_file is not None and output_dir is not None)
    command = ("bigmler execute --code-file " + res_filename(code_file) +
               " --store --to-library --output-dir " + output_dir)
    shell_execute(command, "%s/xx.txt" % output_dir)


def i_create_all_execution_with_io_resources(
    step, code_file=None, output_dir=None, inputs_dec=None,
    outputs_dec=None, inputs=None):
    """Step: I create BigML execution resources from code in file <code_file>
    with inputs <inputs_dec>, outputs <outputs_dec> and inputs <inputs> and
    log results in <output_dir>
    """
    ok_(code_file is not None and output_dir is not None and
        inputs_dec is not None and outputs_dec is not None and
        inputs is not None)
    command = ("bigmler execute --code-file " + res_filename(code_file) +
               " --store --declare-inputs " + res_filename(inputs_dec) +
               " --declare-outputs "+ res_filename(outputs_dec) +
               " --inputs " + res_filename(inputs) +
               " --output-dir " + output_dir)
    shell_execute(command, "%s/xx.txt" % output_dir)


def i_create_from_whizzml_package(step, package_dir=None, output_dir=None):
    """Step: 'I create a BigML whizzml package from <package_dir> and log
    results in <output_dir>
    """
    ok_(package_dir is not None and output_dir is not None)
    command = ("bigmler whizzml --package-dir " + res_filename(package_dir) +
               " --store --output-dir " + output_dir)
    shell_execute(command, "%s/xx.txt" % output_dir)


def i_check_create_script(step):
    """Step: I check that the script has been created"""
    script_file = os.path.join(world.directory, "scripts")
    message = None
    try:
        with open(script_file) as handler:
            script = check_resource(handler.readline().strip(),
                                    world.api.get_script)
            world.scripts.append(script['resource'])
            world.script = script
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_create_execution(step, number_of_executions=1):
    """Step: I check that the execution has been created"""
    execution_file = os.path.join(world.directory, "execution")
    message = None
    try:
        with open(execution_file) as handler:
            for _ in range(number_of_executions):
                execution = check_resource(handler.readline().strip(),
                                           world.api.get_execution)
                world.executions.append(execution['resource'])
                world.execution = execution
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_create_result(step):
    """Step: I check that the result is ready"""
    result_file = os.path.join(world.directory, "whizzml_results.json")
    message = None
    try:
        with open(result_file) as handler:
            world.results = json.load(handler)
        del world.results["sources"]
    except Exception as exc:
        message = str(exc)
    ok_(message is None, msg=message)


def i_check_result_is(step, check_file=None):
    """Step: the result is like <check_file>"""
    try:
        with open(res_filename(check_file)) as handler:
            results = json.load(handler)
        del results["sources"]
        eq_(results, world.results)
    except Exception as exc:
        message = str(exc)
        ok_(message is None, msg=message)


def i_check_create_library(step):
    """Step: I check that the library has been created"""
    library_file = os.path.join(world.directory, "library")
    message = None
    try:
        with open(library_file) as handler:
            library = check_resource(handler.readline().strip(),
                                     world.api.get_library)
            world.libraries.append(library['resource'])
            world.library = library
    except Exception as exc:
        message = str(exc)
        ok_(message is None, msg=message)


def i_check_create_package_script(step, package_dir=None):
    """Step: I check that the package script in <package_dir> has been
    created
    """
    script_file = os.path.join(world.directory, os.path.basename(package_dir),
                               "scripts")
    message = None
    try:
        with open(script_file) as handler:
            script = check_resource(handler.readline().strip(),
                                    world.api.get_script)
            world.scripts.append(script['resource'])
            world.script = script
    except Exception as exc:
        message = str(exc)
        ok_(message is None, msg=message)



def i_create_from_whizzml_package_embedding(step, package_dir=None,
                                            output_dir=None):
    """Step: I create a BigML whizzml package from <package_dir>, embed any
    library and log results in <output_dir>
    """
    ok_(package_dir is not None and output_dir is not None)
    command = ("bigmler whizzml --package-dir " + res_filename(package_dir) +
               " --embed-libs --store --output-dir " + output_dir)
    shell_execute(command, "%s/xx.txt" % output_dir)



# -*- coding: utf-8 -*-
#
# Copyright 2020-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Resources management functions

"""


import sys

import bigml.api

from bigmler.utils import (dated, get_url, log_message, check_resource,
                           plural, is_shared,
                           check_resource_error, log_created_resources)
from bigmler.reports import report

from bigmler.resourcesapi.common import set_basic_model_args, \
    configure_input_fields, update_sample_parameters_args, \
    update_json_args, wait_for_available_tasks

from bigmler.resourcesapi.common import SEED, FIELDS_QS, \
    ALL_FIELDS_QS


def set_cluster_args(args, name=None, fields=None,
                     cluster_fields=None):
    """Return cluster arguments dict

    """
    if name is None:
        name = args.name
    if cluster_fields is None:
        cluster_fields = args.cluster_fields_

    cluster_args = set_basic_model_args(args, name)
    cluster_args.update({
        "seed": SEED if args.seed is None else args.seed,
        "cluster_seed": (SEED if args.cluster_seed is None
                         else args.cluster_seed)
    })

    if args.cluster_models is not None:
        cluster_args.update({"model_clusters": True})
    if args.cluster_k:
        cluster_args.update({"k": args.cluster_k})
    if cluster_fields and fields is not None:
        input_fields = configure_input_fields(fields, cluster_fields)
        cluster_args.update(input_fields=input_fields)
    if args.summary_fields is not None:
        cluster_args.update({"summary_fields": args.summary_fields_})

    cluster_args = update_sample_parameters_args(cluster_args, args)

    if 'cluster' in args.json_args:
        update_json_args(cluster_args, args.json_args.get('cluster'), fields)

    return cluster_args


def create_clusters(datasets, cluster_ids, cluster_args,
                    args, api=None, path=None,
                    session_file=None, log=None):
    """Create remote clusters

    """
    if api is None:
        api = bigml.api.BigML()

    clusters = cluster_ids[:]
    existing_clusters = len(clusters)
    cluster_args_list = []
    datasets = datasets[existing_clusters:]
    # if resuming and all clusters were created, there will be no datasets left
    if datasets:
        if isinstance(cluster_args, list):
            cluster_args_list = cluster_args

        # Only one cluster per command, at present
        number_of_clusters = 1
        message = dated("Creating %s.\n" %
                        plural("cluster", number_of_clusters))
        log_message(message, log_file=session_file,
                    console=args.verbosity)

        query_string = FIELDS_QS
        inprogress = []
        for i in range(0, number_of_clusters):
            wait_for_available_tasks(inprogress, args.max_parallel_clusters,
                                     api, "cluster")
            if cluster_args_list:
                cluster_args = cluster_args_list[i]

            cluster = api.create_cluster(datasets, cluster_args, retries=None)
            cluster_id = check_resource_error(cluster,
                                              "Failed to create cluster: ")
            log_message("%s\n" % cluster_id, log_file=log)
            cluster_ids.append(cluster_id)
            inprogress.append(cluster_id)
            clusters.append(cluster)
            log_created_resources("clusters", path, cluster_id, mode='a')

        if args.verbosity:
            if bigml.api.get_status(cluster)['code'] != bigml.api.FINISHED:
                try:
                    cluster = check_resource(cluster, api.get_cluster,
                                             query_string=query_string,
                                             raise_on_error=True)
                except Exception as exception:
                    sys.exit("Failed to get a finished cluster: %s" %
                             str(exception))
                clusters[0] = cluster
            message = dated("Cluster created: %s\n" %
                            get_url(cluster))
            log_message(message, log_file=session_file,
                        console=args.verbosity)
            if args.reports:
                report(args.reports, path, cluster)

    return clusters, cluster_ids


def get_clusters(cluster_ids, args, api=None, session_file=None):
    """Retrieves remote clusters in its actual status

    """
    if api is None:
        api = bigml.api.BigML()
    cluster_id = ""
    clusters = cluster_ids
    cluster_id = cluster_ids[0]
    message = dated("Retrieving %s. %s\n" %
                    (plural("cluster", len(cluster_ids)),
                     get_url(cluster_id)))
    log_message(message, log_file=session_file, console=args.verbosity)
    # only one cluster to predict at present
    try:
        # we need the whole fields structure when exporting fields
        query_string = FIELDS_QS if not args.export_fields else ALL_FIELDS_QS
        cluster = check_resource(cluster_ids[0], api.get_cluster,
                                 query_string=query_string,
                                 raise_on_error=True)
    except Exception as exception:
        sys.exit("Failed to get a finished cluster: %s" % str(exception))
    clusters[0] = cluster

    return clusters, cluster_ids


def set_publish_cluster_args(args):
    """Set args to publish cluster

    """
    public_cluster = {}
    if args.public_cluster:
        public_cluster = {"private": False}
        if args.model_price:
            public_cluster.update(price=args.model_price)
        if args.cpp:
            public_cluster.update(credits_per_prediction=args.cpp)
    return public_cluster


def update_cluster(cluster, cluster_args, args,
                   api=None, path=None, session_file=None):
    """Updates cluster properties

    """
    if api is None:
        api = bigml.api.BigML()
    message = dated("Updating cluster. %s\n" %
                    get_url(cluster))
    log_message(message, log_file=session_file,
                console=args.verbosity)
    cluster = api.update_cluster(cluster, cluster_args)
    check_resource_error(cluster, "Failed to update cluster: %s"
                         % cluster['resource'])
    cluster = check_resource(cluster, api.get_cluster, query_string=FIELDS_QS,
                             raise_on_error=True)
    if is_shared(cluster):
        message = dated("Shared cluster link. %s\n" %
                        get_url(cluster, shared=True))
        log_message(message, log_file=session_file, console=args.verbosity)
        if args.reports:
            report(args.reports, path, cluster)

    return cluster



# -*- coding: utf-8 -*-
#
# Copyright 2020-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Resources management functions

"""

import sys

import bigml.api

from bigml.multivote import THRESHOLD_CODE

from bigmler.utils import (dated, get_url, log_message, check_resource,
                           check_resource_error, log_created_resources,
                           is_shared)
from bigmler.reports import report

from bigmler.resourcesapi.common import set_basic_args, map_fields, \
    update_json_args, get_basic_seed, wait_for_available_tasks, \
    save_txt_and_json
from bigmler.labels import label_model_args
from bigmler.resourcesapi.common import EVALUATE_SAMPLE_RATE, \
    SEED


def set_evaluation_args(args, fields=None,
                        dataset_fields=None, name=None):
    """Return evaluation args dict

    """
    if name is None:
        name = args.name
    evaluation_args = set_basic_args(args, name)

    if hasattr(args, 'method') and (args.number_of_models > 1
                                    or args.ensemble):
        evaluation_args.update(combiner=args.method)
    if hasattr(args, 'method') and args.method:
        evaluation_args.update({"combiner": args.method})
        if args.method == THRESHOLD_CODE:
            threshold = {}
            if hasattr(args, 'threshold') and args.threshold is not None:
                threshold.update(k=args.threshold)
            if hasattr(args, 'threshold_class') \
                    and args.threshold_class is not None:
                threshold.update({"class": args.threshold_class})
            evaluation_args.update(threshold=threshold)
    if args.fields_map_ and fields is not None:
        if dataset_fields is None:
            dataset_fields = fields
        evaluation_args.update({"fields_map": map_fields(args.fields_map_,
                                                         fields,
                                                         dataset_fields)})
    if hasattr(args, 'missing_strategy') and args.missing_strategy:
        evaluation_args.update(missing_strategy=args.missing_strategy)
    if 'evaluation' in args.json_args:
        update_json_args(
            evaluation_args, args.json_args.get('evaluation'), fields)
    # if evaluating time series we need to use ranges
    if args.subcommand == "time-series" and args.test_split == 0 and \
            not args.has_test_datasets_:
        args.range_ = [int(args.max_rows * EVALUATE_SAMPLE_RATE) + 1,
                       args.max_rows]
        evaluation_args.update({"range": args.range_})
        return evaluation_args
    # Two cases to use out_of_bag and sample_rate: standard evaluations where
    # only the training set is provided, and cross_validation
    # [--dataset|--test] [--model|--models|--model-tag|--ensemble] --evaluate
    if (((hasattr(args, "dataset") and args.dataset) or args.test_set)
            and args.has_supervised_):
        return evaluation_args
    # [--train|--dataset] --test-split --evaluate
    if args.test_split > 0 and (args.training_set or args.dataset):
        return evaluation_args
    # --datasets --test-datasets or equivalents
    #if args.datasets and (args.test_datasets or args.dataset_off):
    if args.has_datasets_ and (args.has_test_datasets_ or args.dataset_off):
        return evaluation_args
    if args.sample_rate == 1:
        args.sample_rate = EVALUATE_SAMPLE_RATE
    evaluation_args.update(out_of_bag=True, seed=SEED,
                           sample_rate=args.sample_rate)
    return evaluation_args


def set_label_evaluation_args(args, labels, all_labels,
                              number_of_evaluations, fields, dataset_fields,
                              objective_field):
    """Set of args needed to build an evaluation per label

    """
    if objective_field is None:
        try:
            objective_id = fields.field_id(fields.objective_field)
        except ValueError as exc:
            sys.exit(exc)
        objective_field = fields.fields[objective_id]['name']
    evaluation_args_list = []

    for index in range(number_of_evaluations - 1, -1, -1):
        label = labels[index]
        new_name = label_model_args(
            args.name, label, all_labels, [], objective_field)[0]
        evaluation_args = set_evaluation_args(args,
                                              fields=fields,
                                              dataset_fields=dataset_fields,
                                              name=new_name)
        evaluation_args_list.append(evaluation_args)
    return evaluation_args_list


def create_evaluations(model_or_ensemble_ids, datasets, evaluation_args,
                       args, api=None,
                       path=None, session_file=None, log=None,
                       existing_evaluations=0):
    """Create evaluations for a list of models

       ``model_or_ensemble_ids``: list of model or ensemble ids to create
                                  an evaluation of
       ``datasets``: dataset objects or ids to evaluate with
       ``evaluation_args``: arguments for the ``create_evaluation`` call
       ``args``: input values for bigmler flags
       ``api``: api to remote objects in BigML
       ``path``: directory to store the BigMLer generated files in
       ``session_file``: file to store the messages of that session
       ``log``: user provided log file
       ``existing_evaluations``: evaluations found when attempting resume
    """

    evaluations = []
    dataset = datasets[0]
    evaluation_args_list = []
    if isinstance(evaluation_args, list):
        evaluation_args_list = evaluation_args
    if api is None:
        api = bigml.api.BigML()
    remaining_ids = model_or_ensemble_ids[existing_evaluations:]
    if args.test_dataset_ids or args.dataset_off:
        remaining_datasets = datasets[existing_evaluations:]
    number_of_evaluations = len(remaining_ids)
    message = dated("Creating evaluations.\n")
    log_message(message, log_file=session_file,
                console=args.verbosity)

    inprogress = []
    for i in range(0, number_of_evaluations):
        model = remaining_ids[i]
        if args.test_dataset_ids or args.dataset_off:
            dataset = remaining_datasets[i]
        wait_for_available_tasks(inprogress, args.max_parallel_evaluations,
                                 api, "evaluation")

        if evaluation_args_list != []:
            evaluation_args = evaluation_args_list[i]
        if args.cross_validation_rate > 0:
            new_seed = get_basic_seed(i + existing_evaluations)
            evaluation_args.update(seed=new_seed)
        evaluation = api.create_evaluation(model, dataset, evaluation_args,
                                           retries=None)
        evaluation_id = check_resource_error(evaluation,
                                             "Failed to create evaluation: ")
        inprogress.append(evaluation_id)
        log_created_resources("evaluations", path, evaluation_id,
                              mode='a')
        evaluations.append(evaluation)
        log_message("%s\n" % evaluation['resource'], log_file=log)

    if (args.number_of_evaluations < 2 and len(evaluations) == 1
            and args.verbosity):
        evaluation = evaluations[0]
        if bigml.api.get_status(evaluation)['code'] != bigml.api.FINISHED:
            try:
                evaluation = check_resource(evaluation, api.get_evaluation,
                                            raise_on_error=True)
            except Exception as exception:
                sys.exit("Failed to get a finished evaluation: %s" %
                         str(exception))
            evaluations[0] = evaluation
        message = dated("Evaluation created: %s\n" %
                        get_url(evaluation))
        log_message(message, log_file=session_file,
                    console=args.verbosity)
        if args.reports:
            report(args.reports, path, evaluation)

    return evaluations


def get_evaluation(evaluation, api=None, verbosity=True, session_file=None):
    """Retrieves evaluation in its actual state

    """
    if api is None:
        api = bigml.api.BigML()
    message = dated("Retrieving evaluation. %s\n" %
                    get_url(evaluation))
    log_message(message, log_file=session_file, console=verbosity)
    try:
        evaluation = check_resource(evaluation, api.get_evaluation,
                                    raise_on_error=True)
    except Exception as exception:
        sys.exit("Failed to get a finished evaluation: %s" % str(exception))
    return evaluation


def update_evaluation(evaluation, evaluation_args, args,
                      api=None, path=None, session_file=None):
    """Updates evaluation properties

    """
    if api is None:
        api = bigml.api.BigML()
    message = dated("Updating evaluation. %s\n" %
                    get_url(evaluation))
    log_message(message, log_file=session_file,
                console=args.verbosity)
    evaluation = api.update_evaluation(evaluation, evaluation_args)
    check_resource_error(evaluation, "Failed to update evaluation: %s"
                         % evaluation['resource'])
    evaluation = check_resource(evaluation, api.get_evaluation,
                                raise_on_error=True)
    if is_shared(evaluation):
        message = dated("Shared evaluation link. %s\n" %
                        get_url(evaluation, shared=True))
        log_message(message, log_file=session_file, console=args.verbosity)
        if args.reports:
            report(args.reports, path, evaluation)

    return evaluation


def save_evaluation(evaluation, output, api=None):
    """Creates the evaluation .txt and .json files

    """
    if api is None:
        api = bigml.api.BigML()
    evaluation = evaluation.get('object', evaluation).get('result', evaluation)
    save_txt_and_json(evaluation, output, api=api)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,invalid-name
#
# Copyright 2014-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

""" Testing batch prediction creation

"""
from bigmler.tests.world import (world, common_setup_module,
                                 common_teardown_module, show_method)

import bigmler.tests.basic_tst_prediction_steps as test_pred
import bigmler.tests.basic_batch_tst_prediction_steps as test_batch_pred
import bigmler.tests.basic_anomaly_prediction_steps as anomaly_pred


def setup_module():
    """Setup for the module

    """
    common_setup_module()
    test = TestBatchPrediction()
    test.bigml = {"method": "setup_scenario2"}
    test.setup_scenario2()

def teardown_module():
    """Teardown for the module

    """
    common_teardown_module()


class TestBatchPrediction:
    """Testing batch prediction commands"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """Calling generic teardown for every method

        """
        world.clear_paths()
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
            Scenario 1: Successfully building test predictions from scratch:
                Given I create BigML resources uploading train "<data>" file to test "<test>" remotely with mapping file "<fields_map>" and log predictions in "<output>"
                And I check that the source has been created
                And I check that the dataset has been created
                And I check that the model has been created
                And I check that the source has been created from the test file
                And I check that the dataset has been created from the test file
                And I check that the batch prediction has been created
                And I check that the predictions are ready
                Then the local prediction file is like "<predictions_file>"

        """
        print(self.test_scenario1.__doc__)
        headers = ["data", "test", "fields_map", "output", "predictions_file"]
        examples = [
            ['data/grades.csv', 'data/test_grades.csv',
             'data/grades_fields_map.csv', 'scenario_r1_r/predictions.csv',
             'check_files/predictions_grades.csv']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            test_pred.i_create_all_resources_batch_map(
                self, data=example["data"], test=example["test"],
                fields_map=example["fields_map"], output=example["output"])
            test_pred.i_check_create_source(self)
            test_pred.i_check_create_dataset(self, suffix=None)
            test_pred.i_check_create_model(self)
            test_batch_pred.i_check_create_test_source(self)
            test_batch_pred.i_check_create_test_dataset(self)
            test_batch_pred.i_check_create_batch_prediction(self)
            test_pred.i_check_create_predictions(self)
            test_pred.i_check_predictions(self, example["predictions_file"])

    def setup_scenario2(self):
        """
            Scenario 2: Successfully building test predictions from scratch:
                Given I create BigML resources uploading train "<data>" file to test "<test>" remotely and log predictions in "<output>"
                And I check that the source has been created
                And I check that the dataset has been created
                And I check that the model has been created
                And I check that the source has been created from the test file
                And I check that the dataset has been created from the test file
                And I check that the batch prediction has been created
                And I check that the predictions are ready
                Then the local prediction file is like "<predictions_file>"
        """
        print(self.setup_scenario2.__doc__)
        headers = ["data", "test", "output", "predictions_file"]
        examples = [
            ['data/iris.csv', 'data/test_iris.csv',
             'scenario_r1/predictions.csv',
             'check_files/predictions_iris.csv']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            test_pred.i_create_all_resources_batch(
                self, data=example["data"], test=example["test"],
                output=example["output"])
            test_pred.i_check_create_source(self)
            test_pred.i_check_create_dataset(self, suffix=None)
            test_pred.i_check_create_model(self)
            test_batch_pred.i_check_create_test_source(self)
            test_batch_pred.i_check_create_test_dataset(self)
            test_batch_pred.i_check_create_batch_prediction(self)
            test_pred.i_check_create_predictions(self)
            test_pred.i_check_predictions(self, example["predictions_file"])

    def test_scenario3(self):
        """
            Scenario 3: Successfully building test predictions from source
                Given I have previously executed "<scenario>" or reproduce it with arguments <kwargs>
                And I create BigML resources using source to test the previous test source remotely and log predictions in "<output>"
                And I check that the dataset has been created
                And I check that the model has been created
                And I check that the dataset has been created from the test file
                And I check that the batch prediction has been created
                And I check that the predictions are ready
                Then the local prediction file is like "<predictions_file>"
        """

        print(self.test_scenario3.__doc__)
        headers = ["scenario", "kwargs", "output", "predictions_file"]
        examples = [
            ['scenario_r1', '{"data": "data/iris.csv",' +
             ' "output": "scenario_r1/predictions.csv",' +
             ' "test": "data/test_iris.csv"}',
             'scenario_r2/predictions.csv',
             'check_files/predictions_iris.csv']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            test_pred.i_have_previous_scenario_or_reproduce_it(
                self, example["scenario"], example["kwargs"])
            test_pred.i_create_resources_from_source_batch(
                self, output=example["output"])
            test_pred.i_check_create_dataset(self, suffix=None)
            test_pred.i_check_create_model(self)
            test_batch_pred.i_check_create_test_dataset(self)
            test_batch_pred.i_check_create_batch_prediction(self)
            test_pred.i_check_create_predictions(self)
            test_pred.i_check_predictions(self, example["predictions_file"])

    def test_scenario4(self):
        """
            Scenario 4: Successfully building test predictions from dataset
                Given I have previously executed "<scenario>" or reproduce it with arguments <kwargs>
                And I create BigML resources using dataset to test the previous test dataset remotely and log predictions in "<output>"
                And I check that the model has been created
                And I check that the batch prediction has been created
                And I check that the predictions are ready
                Then the local prediction file is like "<predictions_file>"
        """

        print(self.test_scenario4.__doc__)
        headers = ["scenario", "kwargs", "output", "predictions_file"]
        examples = [
            ['scenario_r1', '{"data": "data/iris.csv",' +
             ' "output": "scenario_r1/predictions.csv",' +
             ' "test": "data/test_iris.csv"}',
             'scenario_r3/predictions.csv',
             'check_files/predictions_iris.csv']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            test_pred.i_have_previous_scenario_or_reproduce_it(
                self, example["scenario"], example["kwargs"])
            test_pred.i_create_resources_from_dataset_batch(
                self, output=example["output"])
            test_pred.i_check_create_model(self)
            test_batch_pred.i_check_create_batch_prediction(self)
            test_pred.i_check_create_predictions(self)
            test_pred.i_check_predictions(self, example["predictions_file"])

    def test_scenario5(self):
        """
            Scenario 5: Successfully building test predictions from dataset and prediction format info
                Given I have previously executed "<scenario>" or reproduce it with arguments <kwargs>
                And I create BigML resources using a model to test the previous test dataset remotely with prediction headers and fields "<fields>" and log predictions in "<output>"
                And I check that the batch prediction has been created
                And I check that the predictions are ready
                Then the local prediction file is like "<predictions_file>"
        """

        print(self.test_scenario5.__doc__)
        headers = ["scenario", "kwargs", "fields", "output",
            "predictions_file"]
        examples = [
            ['scenario_r1', '{"data": "data/iris.csv",' +
             ' "output": "scenario_r1/predictions.csv",' +
             ' "test": "data/test_iris.csv"}',
             'sepal length,sepal width',
             'scenario_r4/predictions.csv',
             'check_files/predictions_iris_format.csv']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            test_pred.i_have_previous_scenario_or_reproduce_it(
                self, example["scenario"], example["kwargs"])
            test_pred.i_create_resources_from_model_batch(
                self, fields=example["fields"], output=example["output"])
            test_batch_pred.i_check_create_batch_prediction(self)
            test_pred.i_check_create_predictions(self)
            test_pred.i_check_predictions(self, example["predictions_file"])

    def test_scenario6(self):
        """
            Scenario 6: Successfully building remote test predictions from scratch to a dataset:
                Given I create BigML resources uploading train "<data>" file to test "<test>" remotely to a dataset with no CSV output and log resources in "<output_dir>"
                And I check that the source has been created
                And I check that the dataset has been created
                And I check that the model has been created
                And I check that the source has been created from the test file
                And I check that the dataset has been created from the test file
                And I check that the batch prediction has been created
                Then I check that the batch predictions dataset exists
                And no local CSV file is created
        """

        print(self.test_scenario6.__doc__)
        headers = ["data", "test", "output_dir"]
        examples = [
            ['data/iris.csv', 'data/test_iris.csv', 'scenario_r5']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            test_pred.i_create_all_resources_batch_to_dataset(
                self, data=example["data"], test=example["test"],
                output_dir=example["output_dir"])
            test_pred.i_check_create_source(self)
            test_pred.i_check_create_dataset(self, suffix=None)
            test_pred.i_check_create_model(self)
            test_batch_pred.i_check_create_test_source(self)
            test_batch_pred.i_check_create_test_dataset(self)
            test_batch_pred.i_check_create_batch_prediction(self)
            test_batch_pred.i_check_create_batch_predictions_dataset(self)
            anomaly_pred.i_check_no_local_csv(self)

    def test_scenario7(self):
        """
        Scenario: Successfully building test predictions from model with operating point
            Given I have previously executed "<scenario>" or reproduce it with arguments <kwargs>
            And I create BigML resources using model with operating point "<operating_point>" to test "<test>" and log predictions in "<output>"
            And I check that the predictions are ready
            Then the local prediction file is like "<predictions_file>"
        """
        headers = ["scenario", "kwargs", "test", "output",
            "predictions_file", "operating_point"]
        examples = [
            ['scenario_r1', '{"data": "data/iris.csv",' +
             ' "output": "scenario_r1/predictions.csv",' +
             ' "test": "data/test_iris.csv"}',
             'data/test_iris.csv', 'scenario_r7/predictions_p.csv', 'check_files/predictions_iris_op_prob.csv', "data/operating_point_prob.json"],
            ['scenario_r1', '{"data": "data/iris.csv",' +
             ' "output": "scenario_r1/predictions.csv",' +
             ' "test": "data/test_iris.csv"}', 'data/test_iris.csv',
             'scenario_r7/predictions_c.csv',
             'check_files/predictions_iris_op_conf.csv',
             "data/operating_point_conf.json"]]

        print(self.test_scenario7.__doc__)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            test_pred.i_have_previous_scenario_or_reproduce_it(
                self, example["scenario"], example["kwargs"])
            test_batch_pred.i_create_resources_from_model_with_op_remote(
                self, operating_point=example["operating_point"],
                test=example["test"], output=example["output"])
            test_batch_pred.i_check_create_batch_prediction(self)
            test_pred.i_check_create_predictions(self)
            test_pred.i_check_predictions(self, example["predictions_file"],
                                          headers=True)

    def test_scenario8(self):
        """
        Scenario: Successfully building test predictions from model with operating point
            Given I have previously executed "<scenario>" or reproduce it with arguments <kwargs>
            And I create BigML remote batch predictions one by one using model to test "<test>" and log predictions in "<output>"
            And I check that the predictions are ready
            Then the local prediction file is like "<predictions_file>"
        """
        headers = ["scenario", "kwargs", "test", "output",
            "predictions_file"]
        examples = [
            ['scenario_r1', '{"data": "data/iris.csv",' +
             ' "output": "scenario_r1/predictions.csv",' +
             ' "test": "data/test_iris.csv"}', 'data/test_iris.csv',
             'scenario_r8/predictions.csv',
             'check_files/predictions_iris.csv']]

        print(self.test_scenario8.__doc__)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            test_pred.i_have_previous_scenario_or_reproduce_it( \
                self, example["scenario"], example["kwargs"])
            test_batch_pred.i_create_resources_from_model_remote_no_batch( \
                self, test=example["test"], output=example["output"])
            test_pred.i_check_create_predictions(self)
            test_pred.i_check_predictions(self, example["predictions_file"],
                                          headers=True)
